class Function:
    def __call__(self, *args):
        return self.forward(*args)

    def forward(self, *args): raise NotImplementedError(f"forward not implemented for {type(self)}")
    def backward(self, result_tensor): raise RuntimeError(f"backward not implemented for {type(self)}")



from neuroscribe.core._nn.conv import Conv1d, Conv2d, Conv3d


# ********** Loss Functions **********
def mse_loss(predictions, targets):
    return (predictions - targets).square().mean()


def l1_loss(predictions, targets):
    return (predictions - targets).abs().mean()


def binary_cross_entropy(predictions, targets):
    predictions = predictions.clip(min=1e-10, max=1 - 1e-10)
    return (targets * predictions.log() + (1 - targets) * (1 - predictions).log()).mean().neg()


# ********** Activation Functions **********
def relu(input): return input.relu()
def relu6(input): return input.relu6()
def leaky_relu(input, negative_slope=0.01): return input.leaky_relu(negative_slope)
def elu(input, alpha=1.0): return input.elu(alpha)
def swish(input): return input.swish()
def silu(input): return input.silu()
def tanh(input): return input.tanh()
def hardtanh(input, min=-1, max=1): return input.hardtanh(min, max)
def gelu(input): return input.gelu()
def sigmoid(input): return input.sigmoid()
def softmax(input): return input.softmax()
def log_softmax(input): return input.log_softmax()
def softmin(input): return input.softmin()
def softplus(input, beta=1.0): return input.softplus(beta)
def mish(input): return input.mish()
def softsign(input): return input.softsign()


# ********** NN Layers **********
def linear(input, weight, bias):
    if bias is not None:
        return input @ weight + bias
    return input @ weight


# TODO: Add support to dilation and groups
def conv1d(input, weight, bias=None, stride=1, padding=0, dilation=1, groups=1):
    return input._exec_op(Conv1d(weight, bias, stride, padding, dilation, groups))


# TODO: Add support to dilation and groups
def conv2d(input, weight, bias=None, stride=1, padding=0, dilation=1, groups=1):
    return input._exec_op(Conv2d(weight, bias, stride, padding, dilation, groups))


# TODO: Add support to dilation and groups
def conv3d(input, weight, bias=None, stride=1, padding=0, dilation=1, groups=1):
    return input._exec_op(Conv3d(weight, bias, stride, padding, dilation, groups))



import neuroscribe.nn.functional as F
from neuroscribe.core._tensor_lib._tensor import Tensor

from .module import Module


class ReLU(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.relu(input)


class ReLU6(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.relu6(input)


class LeakyReLU(Module):
    def __init__(self, negative_slope=0.01):
        super().__init__()
        self.negative_slope = negative_slope

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.leaky_relu(input, self.negative_slope)


class ELU(Module):
    def __init__(self, alpha=1.0):
        super().__init__()
        self.alpha = alpha

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.elu(input, self.alpha)


class Swish(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.swish(input)


class SiLU(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.silu(input)


class Tanh(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.tanh(input)


class HardTanh(Module):
    def __init__(self, min=-1, max=1):
        super().__init__()
        self.min = min
        self.max = max
        assert self.max > self.min

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.hardtanh(input, self.min, self.max)


class GELU(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.gelu(input)


class Sigmoid(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.sigmoid(input)


class Softmax(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.softmax(input)


class LogSoftmax(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.log_softmax(input)


class Softmin(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.softmin(input)


class Softplus(Module):
    def __init__(self, beta=1.0):
        super().__init__()
        self.beta = beta

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.softplus(input, self.beta)


class Mish(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.mish(input)


class Softsign(Module):
    def __init__(self):
        super().__init__()

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)
        return F.softsign(input)



import math


def calculate_gain(nonlinearity, a):
    if nonlinearity == 'linear' or nonlinearity == 'sigmoid':
        return 1
    elif nonlinearity == 'tanh':
        return 5.0 / 3
    elif nonlinearity == 'relu':
        return math.sqrt(2.0)
    elif nonlinearity == 'leaky_relu':
        return math.sqrt(2.0 / (1 + a**2))
    else:
        raise ValueError(f"Unsupported nonlinearity '{nonlinearity}'")


def calculate_correct_fan(tensor, mode, *, _reverse=False):
    dimensions = tensor.ndim
    if dimensions < 2:
        raise ValueError("Fan in and fan out can not be computed for tensor with fewer than 2 dimensions")
    num_in_feature_maps = tensor.shape[0] if _reverse else tensor.shape[1]
    num_out_feature_maps = tensor.shape[1] if _reverse else tensor.shape[0]
    receptive_field_size = 1
    if dimensions > 2:
        receptive_field_size = math.prod(tensor.shape[2:])
    if mode == 'fan_in':
        return num_in_feature_maps * receptive_field_size
    elif mode == 'fan_out':
        return num_out_feature_maps * receptive_field_size
    else:
        raise ValueError(f"Unsupported mode '{mode}', please use one of ['fan_in', 'fan_out']")


def kaiming_normal(tensor, a=0, mode='fan_in', nonlinearity='leaky_relu', *, _reverse=False):
    fan = calculate_correct_fan(tensor, mode, _reverse=_reverse)
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)

    tensor.normal_(0, std)


def kaiming_uniform(tensor, a=0, mode='fan_in', nonlinearity='leaky_relu', *, _reverse=False):
    fan = calculate_correct_fan(tensor, mode, _reverse=_reverse)
    gain = calculate_gain(nonlinearity, a)
    std = gain / math.sqrt(fan)
    bound = math.sqrt(3.0) * std

    tensor.uniform_(-bound, bound)


def xavier_normal(tensor, gain=1.0, *, _reverse=False):
    fan_in = calculate_correct_fan(tensor, mode='fan_in', _reverse=_reverse)
    fan_out = calculate_correct_fan(tensor, mode='fan_out', _reverse=_reverse)
    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))

    tensor.normal_(0, std)


def xavier_uniform(tensor, gain=1.0, *, _reverse=False):
    fan_in = calculate_correct_fan(tensor, mode='fan_in', _reverse=_reverse)
    fan_out = calculate_correct_fan(tensor, mode='fan_out', _reverse=_reverse)
    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))
    bound = math.sqrt(3.0) * std

    tensor.uniform_(-bound, bound)



import numpy as np
import plotly.graph_objs as go

from neuroscribe.core._tensor_lib._tensor import Tensor
from neuroscribe.core._utils._data import _convert_to_numpy

__all__ = [
    'precision_score',
    'recall_score',
    'f1_score',
    'accuracy_score',
    'confusion_matrix',
    'r_squared',
    'mean_squared_error',
    'mean_absolute_error',
    'root_mean_squared_error'
]


def _validate_inputs(y_true, y_pred, validate_predictions=False):
    y_true, y_pred = _convert_to_numpy(y_true, y_pred)

    if len(y_true) != len(y_pred):
        raise ValueError("Length of y_true and y_pred must be the same")

    if validate_predictions:
        invalid_predictions = set(y_pred.flatten()) - set(y_true.flatten())
        if invalid_predictions:
            raise ValueError(
                f"The following predicted values are not in the set of true classes: {sorted(invalid_predictions)}"
            )

    return y_true, y_pred


# ********** Classification Metrics **********
def precision_score(y_true, y_pred, average='macro', visualize=False, validate_predictions=True):
    y_true, y_pred = _validate_inputs(y_true, y_pred, validate_predictions=validate_predictions)

    if average == 'macro':
        classes = np.unique(y_true)
        precision_scores = []
        for cls in classes:
            tp = np.sum((y_true == cls) & (y_pred == cls))
            fp = np.sum((y_true != cls) & (y_pred == cls))
            precision = tp / (tp + fp) if tp + fp > 0 else 0
            precision_scores.append(precision)
        precision = np.mean(precision_scores)
    elif average == 'micro':
        tp = np.sum(y_true == y_pred)
        fp = np.sum(y_true != y_pred)
        precision = tp / (tp + fp)

    if visualize:
        tp_total = np.sum(y_true == y_pred)
        fp_total = np.sum(y_true != y_pred)
        labels = ['True Positive', 'False Positive']
        values = [tp_total, fp_total]
        fig = go.Figure(data=[go.Pie(labels=labels, values=values, hole=.3)])
        fig.update_layout(title_text='Precision Visualization')
        fig.show()

    return Tensor.create(precision)


def recall_score(y_true, y_pred, average='macro', visualize=False, validate_predictions=True):
    y_true, y_pred = _validate_inputs(y_true, y_pred, validate_predictions=validate_predictions)

    if average == 'macro':
        classes = np.unique(y_true)
        recall_scores = []
        for cls in classes:
            tp = np.sum((y_true == cls) & (y_pred == cls))
            fn = np.sum((y_true == cls) & (y_pred != cls))
            recall = tp / (tp + fn) if tp + fn > 0 else 0
            recall_scores.append(recall)
        recall = np.mean(recall_scores)
    elif average == 'micro':
        tp = np.sum(y_true == y_pred)
        fn = np.sum(y_true != y_pred)
        recall = tp / (tp + fn)

    if visualize:
        tp_total = np.sum(y_true == y_pred)
        fn_total = np.sum(y_true != y_pred)
        labels = ['True Positive', 'False Negative']
        values = [tp_total, fn_total]
        fig = go.Figure(data=[go.Pie(labels=labels, values=values, hole=.3)])
        fig.update_layout(title_text='Recall Visualization')
        fig.show()

    return Tensor.create(recall)


def f1_score(y_true, y_pred, average='macro', visualize=False, validate_predictions=True):
    y_true, y_pred = _validate_inputs(y_true, y_pred, validate_predictions=validate_predictions)

    if average == 'macro':
        classes = np.unique(y_true)
        f1_scores = []
        for cls in classes:
            tp = np.sum((y_true == cls) & (y_pred == cls))
            fp = np.sum((y_true != cls) & (y_pred == cls))
            fn = np.sum((y_true == cls) & (y_pred != cls))
            precision = tp / (tp + fp) if tp + fp > 0 else 0
            recall = tp / (tp + fn) if tp + fn > 0 else 0
            f1 = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0
            f1_scores.append(f1)
        f1 = np.mean(f1_scores)
    elif average == 'micro':
        tp = np.sum(y_true == y_pred)
        fp = np.sum(y_true != y_pred)
        fn = np.sum(y_true != y_pred)
        precision = tp / (tp + fp) if tp + fp > 0 else 0
        recall = tp / (tp + fn) if tp + fn > 0 else 0
        f1 = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

    if visualize:
        tp_total = np.sum(y_true == y_pred)
        fp_total = np.sum(y_true != y_pred)
        fn_total = np.sum(y_true != y_pred)
        labels = ['True Positive', 'False Positive', 'False Negative']
        values = [tp_total, fp_total, fn_total]
        fig = go.Figure(data=[go.Pie(labels=labels, values=values, hole=.3)])
        fig.update_layout(title_text='F1 Score Visualization')
        fig.show()

    return Tensor.create(f1)


def accuracy_score(y_true, y_pred, visualize=False, validate_predictions=True):
    y_true, y_pred = _validate_inputs(y_true, y_pred, validate_predictions=validate_predictions)
    accuracy = np.sum(y_true == y_pred) / len(y_true)

    if visualize:
        correct = np.sum(y_true == y_pred)
        incorrect = np.sum(y_true != y_pred)
        labels = ['Correct', 'Incorrect']
        values = [correct, incorrect]
        fig = go.Figure(data=[go.Pie(labels=labels, values=values, hole=.3)])
        fig.update_layout(title_text='Accuracy Visualization')
        fig.show()

    return Tensor.create(accuracy)


def confusion_matrix(y_true, y_pred, visualize=False, validate_predictions=True):
    y_true, y_pred = _validate_inputs(y_true, y_pred, validate_predictions=validate_predictions)
    classes = np.unique(y_true)
    num_classes = len(classes)

    cm = np.zeros((num_classes, num_classes), dtype=int)
    for i, cls_true in enumerate(classes):
        for j, cls_pred in enumerate(classes):
            cm[i, j] = np.sum((y_true == cls_true) & (y_pred == cls_pred))

    if visualize:
        x_axis = [f'Predicted {cls}' for cls in classes]
        y_axis = [f'True {cls}' for cls in classes]

        fig = go.Figure(data=go.Heatmap(
            z=cm,
            x=x_axis,
            y=y_axis,
            colorscale='Blues',
            text=cm,
            hoverinfo='text'
        ))

        fig.update_layout(title_text='Confusion Matrix')
        fig.show()

    return Tensor.create(cm)


# ********** Regression Metrics **********
def r_squared(y_true, y_pred):
    y_true, y_pred = _validate_inputs(y_true, y_pred)
    ss_res = np.sum((y_true - y_pred) ** 2)
    ss_tot = np.sum((y_true - np.mean(y_true)) ** 2)
    r2 = 1 - (ss_res / ss_tot)

    return Tensor.create(r2)


def mean_squared_error(y_true, y_pred):
    y_true, y_pred = _validate_inputs(y_true, y_pred)
    return Tensor.create(np.mean((y_true - y_pred) ** 2))


def mean_absolute_error(y_true, y_pred):
    y_true, y_pred = _validate_inputs(y_true, y_pred)
    return Tensor.create(np.mean(np.abs(y_true - y_pred)))


def root_mean_squared_error(y_true, y_pred):
    y_true, y_pred = _validate_inputs(y_true, y_pred)
    return Tensor.create(np.sqrt(np.mean((y_true - y_pred) ** 2)))



import random

import numpy as np

__all__ = [
    'Transformation',
    'Rescale',
    'Resize',
    'RandomCrop',
    'CenterCrop',
    'RandomHorizontalFlip',
    'RandomVerticalFlip',
    'RandomRotation',
    'Grayscale',
    'Normalize',
    'Compose'
]


class Transformation:
    def __call__(self):
        raise NotImplementedError(f"__call__ not implemented for {type(self)}")


class Rescale(Transformation):
    def __init__(self, scale):
        self.scale = scale

    def __call__(self, x):
        return x * self.scale


class Resize(Transformation):
    def __init__(self, size):
        self.size = size

    def __call__(self, x):
        return np.resize(x, self.size)


# TODO: This can be done in more general way
class RandomCrop(Transformation):
    def __init__(self, size):
        self.size = size

    def __call__(self, x):
        height, width = x.shape[:2]
        top = np.random.randint(0, height - self.size[0])
        left = np.random.randint(0, width - self.size[1])
        return x[top: top + self.size[0], left: left + self.size[1]]


# TODO: This can be done in more general way
class CenterCrop(Transformation):
    def __init__(self, size):
        self.size = size

    def __call__(self, x):
        height, width = x.shape[:2]
        top = (height - self.size[0]) // 2
        left = (width - self.size[1]) // 2
        return x[top: top + self.size[0], left: left + self.size[1]]


class RandomHorizontalFlip(Transformation):
    def __call__(self, x):
        if random.random() > 0.5:
            return np.fliplr(x)
        return x


class RandomVerticalFlip(Transformation):
    def __call__(self, x):
        if random.random() > 0.5:
            return np.flipud(x)
        return x


class RandomRotation(Transformation):
    def __init__(self, degrees):
        self.degrees = degrees

    def __call__(self, x):
        angle = random.uniform(-self.degrees, self.degrees)
        return np.rot90(x, k=int(angle // 90))


class Grayscale(Transformation):
    def __call__(self, x):
        return np.dot(x[..., :3], [0.2989, 0.5870, 0.1140])


class Normalize(Transformation):
    def __init__(self, mean=(0.5,), std=(0.5,), max_pixel_value=255.0):
        self.mean = np.array(mean)
        self.std = np.array(std)
        self.max_pixel_value = max_pixel_value

    def __call__(self, tensor):
        if not isinstance(tensor, np.ndarray):
            raise TypeError(
                f"Input tensor should be a numpy array. Got {type(tensor)}.")

        if tensor.ndim < 2:
            raise ValueError(
                f"Expected tensor to be a tensor image of size (..., C, ...) or (..., ...). Got tensor.shape = {tensor.shape}"
            )

        mean = self.mean * self.max_pixel_value / 2.0
        std = self.std * self.max_pixel_value

        if np.any(std == 0):
            raise ValueError(
                "std evaluated to zero, leading to division by zero.")

        if mean.ndim == 1:
            mean = mean.reshape(-1, *((1,) * (tensor.ndim - 1)))
        if std.ndim == 1:
            std = std.reshape(-1, *((1,) * (tensor.ndim - 1)))

        tensor = (tensor - mean) / std

        return tensor


class Compose:
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, x):
        for t in self.transforms:
            if isinstance(t, Transformation):
                x = t(x)
            else:
                raise TypeError(
                    f"Invalid transformation: {type(t)}. Expected an instance of 'Transformation'"
                )

        return x



import cupy as cp

import neuroscribe.core._tensor_lib.backend.cuda.mlops as mlops


class CUDABackend:

    device = 'cuda'

    @staticmethod
    def argmax(data, dim):
        return cp.argmax(data, dim)

    @staticmethod
    def argmin(data, dim):
        return cp.argmin(data, dim)

    @staticmethod
    def is_contiguous(data):
        return data.flags['C_CONTIGUOUS']

    @staticmethod
    def make_contiguous(data):
        return cp.ascontiguousarray(data)

    @staticmethod
    def deep_copy(data):
        return data.copy(order='C')

    @staticmethod
    def shallow_copy(data):
        return data.view()

    @staticmethod
    def normal_(mean, standard_deviation, shape):
        return cp.random.normal(mean, standard_deviation, shape)

    @staticmethod
    def uniform_(lower_bound, upper_bound, shape):
        return cp.random.uniform(lower_bound, upper_bound, shape)

    @staticmethod
    def arange(start, stop, step, dtype):
        return cp.arange(start, stop, step, dtype=dtype)

    @staticmethod
    def shuffle_(data):
        cp.random.shuffle(data)

    @staticmethod
    def einsum(subscripts, *inputs):
        return cp.einsum(subscripts, *inputs)

    @staticmethod
    def pad(input, pad, mode, constant_values):
        return cp.pad(input, pad, mode, constant_values=constant_values)

    # ********** Creation Methods **********
    # TODO: should optimize when data is already a cupy.ndarray
    @staticmethod
    def create(data, dtype):
        return cp.array(data, dtype=dtype)

    @staticmethod
    def zeros(shape, dtype):
        return cp.zeros(shape, dtype=dtype)

    @staticmethod
    def zeros_like(input, dtype):
        return cp.zeros_like(input, dtype=dtype)

    @staticmethod
    def ones(shape, dtype):
        return cp.ones(shape, dtype=dtype)

    @staticmethod
    def ones_like(input, dtype):
        return cp.ones_like(input, dtype=dtype)

    @staticmethod
    def randn(*shape, dtype):
        return cp.random.randn(*shape).astype(dtype)

    @staticmethod
    def rand(*shape, dtype):
        return cp.random.rand(*shape).astype(dtype)

    @staticmethod
    def randint(low, high, shape, dtype):
        return cp.random.randint(low, high, shape, dtype=dtype)

    @staticmethod
    def empty(shape, dtype):
        return cp.empty(shape, dtype=dtype)

    @staticmethod
    def empty_like(input, dtype):
        return cp.empty_like(input, dtype=dtype)

    # ********** Shape Manipulation Methods **********
    @staticmethod
    def flatten(data):
        return data.flatten()

    @staticmethod
    def flip(data, dims):
        return cp.flip(data, dims)

    @staticmethod
    def squeeze(data, dim):
        return data.squeeze(dim)

    @staticmethod
    def reshape(data, shape):
        return data.reshape(shape)

    @staticmethod
    def transpose(data, dims):
        return data.transpose(dims)

    @staticmethod
    def split(data, indices_or_sections, dim):
        return cp.split(data, indices_or_sections, dim)

    # ********** Unary ops **********
    @staticmethod
    def mean():
        return mlops.Mean()

    @staticmethod
    def sum():
        return mlops.Sum()

    @staticmethod
    def max():
        return mlops.Max()

    @staticmethod
    def min():
        return mlops.Min()

    @staticmethod
    def relu():
        return mlops.ReLU()

    @staticmethod
    def leaky_relu(negative_slope):
        return mlops.LeakyReLU(negative_slope)

    @staticmethod
    def sigmoid():
        return mlops.Sigmoid()

    @staticmethod
    def square():
        return mlops.Square()

    @staticmethod
    def neg():
        return mlops.Neg()

    @staticmethod
    def clip(min, max):
        return mlops.Clip(min, max)

    @staticmethod
    def sign():
        return mlops.Sign()

    @staticmethod
    def reciprocal():
        return mlops.Reciprocal()

    @staticmethod
    def sqrt():
        return mlops.Sqrt()

    @staticmethod
    def log():
        return mlops.Log()

    @staticmethod
    def exp():
        return mlops.Exp()

    @staticmethod
    def sin():
        return mlops.Sin()

    @staticmethod
    def cos():
        return mlops.Cos()

    @staticmethod
    def tanh():
        return mlops.Tanh()

    # ********** Binary Ops **********
    @staticmethod
    def pow():
        return mlops.Pow()

    @staticmethod
    def add():
        return mlops.Add()

    @staticmethod
    def sub():
        return mlops.Sub()

    @staticmethod
    def mul():
        return mlops.Mul()

    @staticmethod
    def div():
        return mlops.Div()

    @staticmethod
    def matmul():
        return mlops.MatMul()



from neuroscribe.core._tensor_lib._tensor import Tensor

from .optimizer import Optimizer


class SGD(Optimizer):
    def __init__(self, params, lr=0.001, momentum=0):
        super().__init__(params)
        if lr < 0:
            raise ValueError(f"Invalid learning rate: {lr}")
        self.lr = lr
        if momentum < 0:
            raise ValueError(f"Invalid momentum value: {momentum}")
        self.momentum = momentum
        self.velocities = [Tensor.zeros_like(param) for param in self.params] if self.momentum != 0 else None

    def step(self):
        if self.momentum == 0:
            for param in self.params:
                param.data = param.data - self.lr * param.grad.data
        else:
            for i, param in enumerate(self.params):
                self.velocities[i].data = self.momentum * self.velocities[i].data + param.grad.data
                param.data = param.data - self.lr * self.velocities[i].data



from neuroscribe.core._tensor_lib._tensor import Tensor

from .function import Function


def execute(operation, *inputs, reverse=False):
    if not isinstance(operation, Function):
        raise TypeError("The 'operation' argument must be an instance of 'Function'.")
    if not inputs:
        raise ValueError("The 'inputs' argument cannot be empty. Provide at least one input.")

    inputs = list(inputs)
    for input in inputs:
        if isinstance(input, Tensor):
            inputs.remove(input)
            return input._exec_op(operation, *inputs, reverse=reverse)

    return Tensor.create(inputs[0])._exec_op(operation, *inputs[1::], reverse=reverse)



from neuroscribe.core._tensor_lib._tensor import Tensor


def einsum(subscripts, *inputs):
    if any(not isinstance(input, Tensor) for input in inputs):
        raise ValueError("All inputs must be of type 'Tensor'")
    inputs_data = tuple(input.data for input in inputs)
    return Tensor(
        inputs[0]._backend.einsum(subscripts, *inputs_data),
        backend=inputs[0]._backend,
        requires_grad=inputs[0].requires_grad,
    )



from neuroscribe.core._tensor_lib._tensor import Tensor


class Optimizer:
    def __init__(self, params):
        self.params = [params] if isinstance(params, Tensor) else list(params)
        if not all(isinstance(param, Tensor) for param in self.params):
            raise TypeError("The params argument must be a Tensor or an iterable of Tensors")
        if len(self.params) == 0:
            raise ValueError("Optimizer received an empty parameter list")

    def zero_grad(self):
        for param in self.params:
            param.grad.zero_()

    def step(self, *args): raise NotImplementedError



def _align_gradient_shape(parameter):
    diff = parameter.grad.ndim - parameter.ndim
    if diff > 0:
        expanded_parameter_shape = (1,) * diff + parameter.shape
        for axis, shape_tuple in enumerate(zip(expanded_parameter_shape, parameter.grad.shape)):
            if shape_tuple[0] == 1 and shape_tuple[1] != 1:
                parameter.grad.data = parameter.grad.data.sum(axis=axis)


def _build_graph(tensor, graph, visited):
    if tensor not in visited:
        visited.add(tensor)
        for child in tensor._prev:
            if child.grad is None:
                raise RuntimeError('Gradient computation has not been enabled for one or more tensor(s).')
            _build_graph(child, graph, visited)
        graph.append(tensor)



import random
from collections import Counter

import numpy as np
import plotly.graph_objects as go

from neuroscribe.core._utils._data import _ensure_ndim

__all__ = [
    'plot_bar',
    'plot_pie',
    'plot_scatter',
    'plot_line',
    'plot_histogram',
    'plot_box',
    'plot_violin',
    'plot_area',
    'plot_bubble',
    'plot_stacked_bar',
    'plot_accuracy',
    'plot_loss',
    'plot_count',
    'plot_feature_correlation',
    'plot_density',
    'plot_heatmap',
    'plot_hexbin',
    'plot_polar_scatter',
    'plot_sunburst',
    'plot_treemap',
    'plot_funnel',
    'plot_word_cloud',
    'plot_3d_scatter',
    'plot_3d_line',
    'plot_3d_surface',
    'plot_3d_mesh'
]


def plot_bar(x, y, title="Bar", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    fig = go.Figure(data=[go.Bar(x=x, y=y)])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_pie(labels, values, title="Pie"):
    labels = _ensure_ndim(labels, ndim=1)
    values = _ensure_ndim(values, ndim=1)
    fig = go.Figure(data=[go.Pie(labels=labels, values=values)])
    fig.update_layout(title=title)
    fig.show()


def plot_scatter(x, y, title="Scatter", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    fig = go.Figure(data=[go.Scatter(x=x, y=y, mode='markers')])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_line(x, y, title="Line", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    fig = go.Figure(data=[go.Scatter(x=x, y=y, mode='lines')])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_histogram(data, nbins=10, title="Histogram", xaxis_title="X Axis", yaxis_title="Y Axis"):
    data = _ensure_ndim(data, ndim=1)
    fig = go.Figure(data=[go.Histogram(x=data, nbinsx=nbins)])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_box(data, labels=None, title="Box"):
    data = [_ensure_ndim(d, ndim=1) for d in data]
    fig = go.Figure()
    if labels is not None:
        labels = _ensure_ndim(labels, ndim=1)
        for d, label in zip(data, labels):
            fig.add_trace(go.Box(y=d, name=label))
    else:
        fig.add_trace(go.Box(y=data))
    fig.update_layout(title=title)
    fig.show()


def plot_violin(data, labels=None, title="Violin"):
    data = [_ensure_ndim(d, ndim=1) for d in data]
    fig = go.Figure()
    if labels is not None:
        labels = _ensure_ndim(labels, ndim=1)
        for d, label in zip(data, labels):
            fig.add_trace(go.Violin(y=d, name=label, box_visible=True, meanline_visible=True))
    else:
        fig.add_trace(go.Violin(y=data, box_visible=True, meanline_visible=True))
    fig.update_layout(title=title)
    fig.show()


def plot_area(x, y, title="Area", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    fig = go.Figure(data=[go.Scatter(x=x, y=y, fill='tozeroy')])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_bubble(x, y, size, title="Bubble", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    size = _ensure_ndim(size, ndim=1)
    fig = go.Figure(data=[go.Scatter(x=x, y=y, mode='markers', marker=dict(size=size))])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_stacked_bar(x, y_data, labels, title="Stacked Bar", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y_data = [_ensure_ndim(y, ndim=1) for y in y_data]
    labels = _ensure_ndim(labels, ndim=1)
    fig = go.Figure()
    for y, label in zip(y_data, labels):
        fig.add_trace(go.Bar(x=x, y=y, name=label))
    fig.update_layout(barmode='stack', title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_accuracy(train_acc, val_acc=None, epochs=None, title="Model Accuracy"):
    train_acc = _ensure_ndim(train_acc, ndim=1)
    if val_acc is not None:
        val_acc = _ensure_ndim(val_acc, ndim=1)
    if epochs is None:
        epochs = np.arange(1, len(train_acc) + 1)
    else:
        epochs = _ensure_ndim(epochs, ndim=1)

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=epochs, y=train_acc, mode='lines', name='Train Accuracy'))
    if val_acc is not None:
        fig.add_trace(go.Scatter(x=epochs, y=val_acc, mode='lines', name='Validation Accuracy'))
    fig.update_layout(title=title, xaxis_title="Epoch", yaxis_title="Accuracy")
    fig.show()


def plot_loss(train_loss, val_loss=None, epochs=None, title="Model Loss"):
    train_loss = _ensure_ndim(train_loss, ndim=1)
    if val_loss is not None:
        val_loss = _ensure_ndim(val_loss, ndim=1)
    if epochs is None:
        epochs = np.arange(1, len(train_loss) + 1)
    else:
        epochs = _ensure_ndim(epochs, ndim=1)

    fig = go.Figure()
    fig.add_trace(go.Scatter(x=epochs, y=train_loss, mode='lines', name='Train Loss'))
    if val_loss is not None:
        fig.add_trace(go.Scatter(x=epochs, y=val_loss, mode='lines', name='Validation Loss'))
    fig.update_layout(title=title, xaxis_title="Epoch", yaxis_title="Loss")
    fig.show()


def plot_count(data, title="Count", xaxis_title="Category", yaxis_title="Count"):
    data = _ensure_ndim(data, ndim=1)
    unique, counts = np.unique(data, return_counts=True)
    fig = go.Figure(data=[go.Bar(x=unique, y=counts)])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_feature_correlation(data, title="Feature Correlation"):
    data = _ensure_ndim(data, ndim=2)
    corr_matrix = np.corrcoef(data, rowvar=False)
    fig = go.Figure(data=go.Heatmap(
        z=corr_matrix,
        x=[f'Feature {i}' for i in range(data.shape[1])],
        y=[f'Feature {i}' for i in range(data.shape[1])],
        colorscale='Viridis'
    ))
    fig.update_layout(title=title)
    fig.show()


def plot_density(data, title="Density", xaxis_title="X Axis", yaxis_title="Density"):
    data = _ensure_ndim(data, ndim=1)
    fig = go.Figure(go.Histogram(
        x=data,
        histnorm='probability',
        marker_color='rgba(0, 0, 255, 0.7)'
    ))
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_heatmap(data, x_labels=None, y_labels=None, title="Heatmap", annotated=False):
    data = _ensure_ndim(data, ndim=2)

    if annotated:
        fig = go.Figure(data=go.Heatmap(
            z=data,
            x=x_labels,
            y=y_labels,
            colorscale='Viridis',
            zmin=np.min(data),
            zmax=np.max(data),
            colorbar=dict(title='Value')
        ))
    else:
        fig = go.Figure(data=go.Heatmap(
            z=data,
            x=x_labels,
            y=y_labels,
            colorscale='Viridis'
        ))

    fig.update_layout(title=title)
    fig.show()


def plot_hexbin(x, y, title="Hexbin", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    fig = go.Figure(data=[go.Histogram2dContour(x=x, y=y, colorscale='Viridis',
                    reversescale=True, contours=dict(coloring='heatmap'))])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_polar_scatter(theta, r, title="Polar Scatter"):
    theta = _ensure_ndim(theta, ndim=1)
    r = _ensure_ndim(r, ndim=1)
    fig = go.Figure(data=[go.Scatterpolar(theta=theta, r=r, mode='markers')])
    fig.update_layout(title=title)
    fig.show()


def plot_sunburst(labels, parents, values=None, title="Sunburst"):
    labels = _ensure_ndim(labels, ndim=1)
    parents = _ensure_ndim(parents, ndim=1)
    fig = go.Figure(data=[go.Sunburst(labels=labels, parents=parents, values=values)])
    fig.update_layout(title=title)
    fig.show()


def plot_treemap(labels, parents, values=None, title="Treemap"):
    labels = _ensure_ndim(labels, ndim=1)
    parents = _ensure_ndim(parents, ndim=1)
    fig = go.Figure(data=[go.Treemap(labels=labels, parents=parents, values=values)])
    fig.update_layout(title=title)
    fig.show()


def plot_funnel(y, x, title="Funnel", xaxis_title="X Axis", yaxis_title="Y Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    fig = go.Figure(data=[go.Funnel(y=y, x=x)])
    fig.update_layout(title=title, xaxis_title=xaxis_title, yaxis_title=yaxis_title)
    fig.show()


def plot_word_cloud(text, title="Word Cloud"):
    words = text.split()
    word_freq = Counter(words)

    max_freq = max(word_freq.values())
    word_freq = {word: freq / max_freq for word, freq in word_freq.items()}

    positions = []
    sizes = []
    for word, freq in word_freq.items():
        x, y = random.uniform(0, 1), random.uniform(0, 1)
        size = freq * 50
        positions.append((word, x, y, size))

    fig = go.Figure()
    for word, x, y, size in positions:
        fig.add_trace(go.Scatter(
            x=[x], y=[y],
            mode='text',
            text=[word],
            textfont=dict(size=size),
            hoverinfo='text',
            textposition='middle center'
        ))

    fig.update_layout(
        title=title,
        xaxis=dict(showgrid=False, zeroline=False, visible=False),
        yaxis=dict(showgrid=False, zeroline=False, visible=False),
        showlegend=False,
        width=800,
        height=400,
        margin=dict(l=0, r=0, t=40, b=0)
    )

    fig.show()


def plot_3d_scatter(x, y, z, title="3D Scatter", xaxis_title="X Axis", yaxis_title="Y Axis", zaxis_title="Z Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    z = _ensure_ndim(z, ndim=1)
    fig = go.Figure(data=[go.Scatter3d(x=x, y=y, z=z, mode='markers')])
    fig.update_layout(title=title, scene=dict(xaxis_title=xaxis_title, yaxis_title=yaxis_title, zaxis_title=zaxis_title))
    fig.show()


def plot_3d_line(x, y, z, title="3D Line", xaxis_title="X Axis", yaxis_title="Y Axis", zaxis_title="Z Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    z = _ensure_ndim(z, ndim=1)
    fig = go.Figure(data=[go.Scatter3d(x=x, y=y, z=z, mode='lines')])
    fig.update_layout(title=title, scene=dict(xaxis_title=xaxis_title, yaxis_title=yaxis_title, zaxis_title=zaxis_title))
    fig.show()


def plot_3d_surface(x, y, z, title="3D Surface", xaxis_title="X Axis", yaxis_title="Y Axis", zaxis_title="Z Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    z = _ensure_ndim(z, ndim=2)
    fig = go.Figure(data=[go.Surface(x=x, y=y, z=z)])
    fig.update_layout(title=title, scene=dict(xaxis_title=xaxis_title, yaxis_title=yaxis_title, zaxis_title=zaxis_title))
    fig.show()


def plot_3d_mesh(x, y, z, title="3D Mesh", xaxis_title="X Axis", yaxis_title="Y Axis", zaxis_title="Z Axis"):
    x = _ensure_ndim(x, ndim=1)
    y = _ensure_ndim(y, ndim=1)
    z = _ensure_ndim(z, ndim=1)
    fig = go.Figure(data=[go.Mesh3d(x=x, y=y, z=z)])
    fig.update_layout(title=title, scene=dict(xaxis_title=xaxis_title, yaxis_title=yaxis_title, zaxis_title=zaxis_title))
    fig.show()



import os

from neuroscribe.core._utils._data import (_read_delimited_data,
                                           _read_ubyte_data,
                                           _read_ubyte_label_data)


def read_data(file_path, chunk_size=1000):
    file_extension = os.path.splitext(file_path)[1:]
    if file_extension in ['.csv', '.tsv']:
        return _read_delimited_data(file_path, chunk_size, delimiter=',' if file_extension == '.csv' else '\t')
    else:
        try:
            with open(file_path, 'rb') as f:
                magic = f.read(4)
                if magic == b'\x00\x00\x08\x03':
                    return _read_ubyte_data(file_path, chunk_size)
                elif magic == b'\x00\x00\x08\x01':
                    return _read_ubyte_label_data(file_path, chunk_size)
                else:
                    raise ValueError("Unsupported file format")
        except FileNotFoundError:
            raise ValueError("File not found")
        except Exception as e:
            raise ValueError("Error reading file: {}".format(e))



from collections import OrderedDict

from .module import Module


class Sequential(Module):
    def __init__(self, *args):
        super().__init__()
        if len(args) == 1 and isinstance(args[0], OrderedDict):
            for key, module in args[0].items():
                self.add_module(key, module)
        else:
            for idx, module in enumerate(args):
                self.add_module(str(idx), module)

    def forward(self, input):
        for module in self._modules.values():
            input = module(input)
        return input



import math

import neuroscribe.nn.functional as F
from neuroscribe.core._tensor_lib._tensor import Tensor

from .. import init
from .module import Module


class Linear(Module):
    def __init__(self, in_features, out_features, bias=True):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = Tensor.empty((in_features, out_features))
        self.bias = Tensor.empty(out_features) if bias else None
        self.init_parameters()

    def init_parameters(self):
        init.kaiming_uniform(self.weight, a=math.sqrt(5), _reverse=True)
        if self.bias is not None:
            fan_in = init.calculate_correct_fan(self.weight, mode='fan_in', _reverse=True)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            self.bias.uniform_(-bound, bound)

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)

        return F.linear(input, self.weight, self.bias)



from neuroscribe.core._tensor_lib._tensor import Tensor

__all__ = ['tensor', 'arange', 'shuffle_', 'zeros', 'zeros_like', 'ones', 'ones_like', 'randn', 'rand', 'randint', 'empty', 'empty_like', 'argmax', 'argmin',
           'add', 'sub', 'mul', 'div', 'matmul', 'relu', 'relu6', 'leaky_relu', 'tanh', 'sinh', 'cosh', 'hardtanh', 'sigmoid', 'softmax', 'log_softmax',
           'softmin', 'mean', 'sum', 'max', 'min', 'square', 'neg', 'clip', 'sign', 'abs', 'reciprocal', 'sqrt', 'rsqrt', 'log', 'log10', 'log2', 'log1p',
           'exp', 'exp2', 'sin', 'cos', 'tan', 'atanh', 'asinh', 'acosh', 'gelu', 'elu', 'swish', 'silu', 'softplus', 'mish', 'softsign', 'pow', 'pad',
           'flatten', 'flip', 'squeeze', 'reshape', 'transpose', 'split']

# Tensor Static Methods
tensor = Tensor.create
arange = Tensor.arange
shuffle_ = Tensor.shuffle_
zeros = Tensor.zeros
zeros_like = Tensor.zeros_like
ones = Tensor.ones
ones_like = Tensor.ones_like
randn = Tensor.randn
rand = Tensor.rand
randint = Tensor.randint
empty = Tensor.empty
empty_like = Tensor.empty_like
pad = Tensor.pad


# Tensor Instance Methods
def flatten(input): return input.flatten()
def flip(input, dims=None): return input.flip(dims)
def squeeze(input, dim=None): return input.squeeze(dim)
def reshape(input, shape): return input.reshape(shape)
def transpose(input, dims=None): return input.transpose(dims)
def split(input, indices_or_sections, dim=0): return input.split(indices_or_sections, dim)
def argmax(input, dim=None): return input.argmax(dim)
def argmin(input, dim=None): return input.argmin(dim)
def relu(input): return input.relu()
def relu6(input): return input.relu6()
def leaky_relu(input, negative_slope=0.01): return input.leaky_relu(negative_slope)
def elu(input, alpha=1.0): return input.elu(alpha)
def swish(input): return input.swish()
def silu(input): return input.silu()
def sigmoid(input): return input.sigmoid()
def softmax(input): return input.softmax()
def log_softmax(input): return input.log_softmax()
def softmin(input): return input.softmin()
def mean(input): return input.mean()
def sum(input): return input.sum()
def max(input): return input.max()
def min(input): return input.min()
def square(input): return input.square()
def neg(input): return input.neg()
def clip(input, min, max): return input.clip(min, max)
def sign(input): return input.sign()
def abs(input): return input.abs()
def reciprocal(input): return input.reciprocal()
def sqrt(input): return input.sqrt()
def rsqrt(input): return input.rsqrt()
def log(input): return input.log()
def log10(input): return input.log10()
def log2(input): return input.log2()
def log1p(input): return input.log1p()
def exp(input): return input.exp()
def exp2(input): return input.exp2()
def sin(input): return input.sin()
def cos(input): return input.cos()
def tan(input): return input.tan()
def tanh(input): return input.tanh()
def sinh(input): return input.sinh()
def cosh(input): return input.cosh()
def atanh(input): return input.atanh()
def asinh(input): return input.asinh()
def acosh(input): return input.acosh()
def hardtanh(input, min=-1, max=1): return input.hardtanh(min, max)
def gelu(input): return input.gelu()
def softplus(input, beta=1.0): return input.softplus(beta)
def mish(input): return input.mish()
def softsign(input): return input.softsign()
def pow(input, exponent): return input.pow(exponent)
def add(input, other): return input.add(other)
def sub(input, other): return input.sub(other)
def mul(input, other): return input.mul(other)
def div(input, other): return input.div(other)
def matmul(input, other): return input.matmul(other)



from neuroscribe.autodiff.executor import execute
from neuroscribe.autodiff.function import Function
from neuroscribe.core._tensor_lib import *
from neuroscribe.nn.modules.module import load, save
from neuroscribe.version import __git_revision__, __version__

__all__ = ['Tensor', 'tensor', 'arange', 'shuffle_', 'zeros', 'zeros_like', 'ones', 'ones_like', 'randn', 'empty', 'empty_like', 'save',
           'load', 'add', 'sub', 'mul', 'matmul', 'relu', 'leaky_relu', 'mean', 'square', 'pad', 'Function', 'execute']



import math

import neuroscribe.nn.init as init
from neuroscribe.nn.modules.module import Module

from .utils import activation_function_map, conv_layer_map


class _ConvND(Module):
    def __init__(self, in_channels, filters, kernel_size, stride=1, padding='valid', dilation=1, groups=1, activation=None, use_bias=None, initializer='kaiming_uniform', dimension=1):
        super().__init__()
        # NOTE: by default the Conv layers is initialized using kaiming_uniform so no further steps are needed
        conv_class = conv_layer_map.get(dimension, None)
        if conv_class is None:
            raise ValueError(f"Invalid dimension argument '{dimension}'. Expected one of [1, 2, 3]")
        self.layer = conv_class(
            in_channels=in_channels,
            out_channels=filters,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            padding_mode='zeros',
            bias=use_bias
        )

        if not isinstance(initializer, str):
            raise TypeError(f"Invalid initializer argument: {type(initializer)}. Expected a string.")
        if initializer not in ['kaiming_uniform', 'kaiming_normal', 'xavier_uniform', 'xavier_normal']:
            raise ValueError(f"Invalid/Unsupported initializer '{initializer}'")
        fan_in = init.calculate_correct_fan(self.layer.weight, mode='fan_in', _reverse=True)
        if initializer.lower() == 'kaiming_normal':
            init.kaiming_normal(self.layer.weight, a=math.sqrt(5), _reverse=True)
            if self.layer.bias:
                std = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
                self.bias.normal_(0, std)
        elif initializer.lower() == 'xavier_uniform':
            init.xavier_uniform(self.layer.weight, _reverse=True)
            if self.layer.bias:
                bound = 1 / math.sqrt(fan_in)
                self.layer.bias.uniform_(-bound, bound)
        elif initializer.lower() == 'xavier_normal':
            init.xavier_normal(self.layer.weight, _reverse=True)
            if self.layer.bias:
                std = 1 / math.sqrt(fan_in)
                self.layer.bias.normal_(0, std)

        if activation:
            if isinstance(activation, str):
                self.activation = activation_function_map.get(activation.lower(), None)
                if self.activation is None:
                    raise ValueError(f"Invalid/Unsupported activation function '{activation}'")
            elif isinstance(activation, Module):
                self.activation = activation
            else:
                raise TypeError(
                    f"Invalid activation argument: {type(activation)}. Expected a string or an instance of 'Module'."
                )
        else:
            self.activation = None

    def forward(self, input):
        output = self.layer(input)
        if self.activation:
            output = self.activation(output)
        return output


class Conv1D(_ConvND):
    def __init__(self, in_channels, filters, kernel_size, stride=1, padding='valid', dilation=1, groups=1, activation=None, use_bias=None, initializer='kaiming_uniform'):
        super().__init__(
            in_channels,
            filters,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            activation,
            use_bias,
            initializer,
            dimension=1
        )


class Conv2D(_ConvND):
    def __init__(self, in_channels, filters, kernel_size, stride=1, padding='valid', dilation=1, groups=1, activation=None, use_bias=None, initializer='kaiming_uniform'):
        super().__init__(
            in_channels,
            filters,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            activation,
            use_bias,
            initializer,
            dimension=2
        )


class Conv3D(_ConvND):
    def __init__(self, in_channels, filters, kernel_size, stride=1, padding='valid', dilation=1, groups=1, activation=None, use_bias=None, initializer='kaiming_uniform'):
        super().__init__(
            in_channels,
            filters,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            activation,
            use_bias,
            initializer,
            dimension=3
        )



from neuroscribe.nn.modules.module import Module
from neuroscribe.optim.optimizer import Optimizer
from neuroscribe.utils.data.data_loader import DataLoader

from .utils import loss_function_map, metric_map, optimizer_map


class Sequential(Module):
    def __init__(self, *args):
        super().__init__()
        for module in args:
            self.add(module)

    def forward(self, input):
        for module in self._modules.values():
            input = module(input)
        return input

    def add(self, module):
        name = self._get_name_for_module(module)
        self.add_module(name, module)

    def _get_name_for_module(self, module):
        type_name = type(module).__name__.lower()
        index = sum(1 for _ in filter(lambda x: type(x) == type(module), self._modules.values()))
        return f'{type_name}_{index}'

    def compile(self, optimizer, loss, metrics=None):
        if isinstance(optimizer, str):
            optimizer_class = optimizer_map.get(optimizer.lower(), None)
            if optimizer_class is None:
                raise ValueError(f"Invalid/Unsupported optimizer '{optimizer}'")
            elif optimizer.lower() == 'sgd_momentum':
                self.optimizer = optimizer_class(self.parameters(), momentum=0.9)
            else:
                self.optimizer = optimizer_class(self.parameters())
        elif isinstance(optimizer, Optimizer):
            self.optimizer = optimizer
        else:
            raise TypeError(
                f"Invalid optimizer argument: {type(optimizer)}. Expected a string or an instance of a subclass of 'Optimizer'."
            )

        if isinstance(loss, str):
            loss_class = loss_function_map.get(loss.lower(), None)
            if loss_class is None:
                raise ValueError(f"Invalid/Unsupported loss functions '{loss}'")
            else:
                self.__dict__['loss'] = loss_class()  # NOTE: To avoid adding the loss function to the _modules ordered dict
        elif isinstance(loss, Module):
            self.__dict__['loss'] = loss  # NOTE: To avoid adding the loss function to the _modules ordered dict
        else:
            raise TypeError(f"Invalid loss argument: {type(loss)}. Expected a string or an instance of a subclass of 'Module'.")

        if metrics is not None:
            if not isinstance(metrics, list):
                raise ("metrics argument must be a list of metric names")

            self.metrics = []
            for metric in metrics:
                metric_fn = metric_map.get(metric, None)
                if metric_fn is None:
                    raise ValueError(f"Invalid/Unsupported metric '{metric}'")
                self.metrics.append(metric_fn)

    def fit(self, train_data, epochs=1, validation_data=None, verbose=False):
        if not hasattr(self, 'optimizer') or not hasattr(self, 'loss'):
            raise RuntimeError("Model must be compiled before training. Use 'compile' method.")
        if not isinstance(train_data, DataLoader):
            raise TypeError(f"Invalid train_data argument: {type(train_data)}. Expected an instance of 'DataLoader'.")

        for epoch in range(epochs):
            self.train()
            total_loss = 0
            num_batches = 0
            verbose_output = ""

            for x, y in train_data:
                self.optimizer.zero_grad()
                outputs = self(x)
                loss = self.loss(outputs, y)
                loss.backward()
                self.optimizer.step()

                total_loss += loss.item()
                num_batches += 1

            avg_loss = total_loss / num_batches

            if verbose:
                verbose_output += f"Epoch {epoch + 1}/{epochs}, Loss: {avg_loss}"

            if validation_data is not None:
                if not isinstance(validation_data, DataLoader):
                    raise TypeError(
                        f"Invalid validation_data argument: {type(validation_data)}. Expected an instance of 'DataLoader'."
                    )
                val_loss, val_metrics = self.evaluate(validation_data, verbose=0)
                if verbose:
                    verbose_output += f", Val Loss: {val_loss}"
                    for name, value in val_metrics.items():
                        verbose_output += f", Val {name}: {value.item()}"

            if verbose_output:
                print(verbose_output)

    def evaluate(self, data, verbose=False):
        if not hasattr(self, 'loss'):
            raise RuntimeError("Model must be compiled before evaluation. Use 'compile' method.")
        if not isinstance(data, DataLoader):
            raise TypeError(f"Invalid data argument: {type(data)}. Expected an instance of 'DataLoader'.")

        self.eval()
        total_loss = 0
        num_batches = 0
        verbose_output = ""
        metrics = {metric_fn.__name__: 0 for metric_fn in self.metrics} if hasattr(self, 'metrics') else {}

        for x, y in data:
            outputs = self(x)
            loss = self.loss(outputs, y)

            total_loss += loss.item()
            num_batches += 1

            if hasattr(self, 'metrics'):
                for metric_fn in self.metrics:
                    metric_name = metric_fn.__name__
                    metrics[metric_name] += metric_fn(y, outputs)

        avg_loss = total_loss / num_batches

        for metric_name in metrics:
            metrics[metric_name] /= num_batches

        if verbose:
            verbose_output += f"Evaluation - Loss: {avg_loss}"
            for name, value in metrics.items():
                verbose_output += f", {name}: {value.item()}"
            print(verbose_output)

        return avg_loss, metrics

    def predict(self, data):
        self.eval()
        predictions = []

        for x in data:
            outputs = self(x)
            predictions.append(outputs)

        return predictions



import os
from concurrent.futures import ThreadPoolExecutor, as_completed

import neuroscribe as ns

__all__ = ['DataLoader']


class DataLoader:
    def __init__(self, dataset, batch_size=None, shuffle=True, train=False, num_workers=0):
        if batch_size is not None and (not isinstance(batch_size, int) or batch_size <= 0):
            raise ValueError("Batch size must be a positive integer or None")
        if not isinstance(num_workers, int) or num_workers < 0:
            raise ValueError(
                "Number of workers must be a non-negative integer")
        if num_workers > os.cpu_count():
            raise ValueError(
                "Number of workers cannot exceed the number of CPU cores")

        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.num_workers = num_workers
        self.indices = ns.arange(len(dataset))
        self.train = train
        self.device = 'cpu'

    def to(self, device):
        self.device = device
        self.dataset.to(device)

    def __len__(self):
        if self.batch_size is None:
            return 1
        else:
            batch_size = self.batch_size or self.indices.numel()
            return (self.indices.numel() + batch_size - 1) // batch_size

    def __iter__(self):
        if self.shuffle:
            ns.shuffle_(self.indices)

        if self.batch_size is None:
            batch_size = self.indices.numel()
        else:
            batch_size = self.batch_size

        if self.num_workers > 0:
            with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
                futures = [executor.submit(self._load_batch, self.indices.data[i:i + batch_size])
                           for i in range(0, self.indices.numel(), batch_size)]
                for future in as_completed(futures):
                    batch = future.result()
                    if batch:
                        yield self._group_batch(batch)
        else:
            for i in range(0, self.indices.numel(), batch_size):
                batch = self._load_batch(self.indices.data[i:i + batch_size])
                if batch:
                    yield self._group_batch(batch)

    def _group_batch(self, batch):
        if not batch:
            return None

        if isinstance(batch[0], tuple):
            num_elements = len(batch[0])
            grouped = tuple(ns.tensor([item[i].data for item in batch], device=self.device,
                            requires_grad=self.train) for i in range(num_elements))
            return grouped

        elif isinstance(batch[0], dict):
            batch_dict = {}
            for key in batch[0].keys():
                batch_dict[key] = ns.tensor(
                    [item[key] for item in batch], requires_grad=self.train)
            return batch_dict

    def _load_batch(self, batch_indices):
        return [self.dataset[idx] for idx in batch_indices if (item := self._load_item(idx)) is not None]

    def _load_item(self, idx):
        try:
            return self.dataset[idx]
        except Exception as e:
            print(f"Error loading item {idx}: {e}")
            return None



import math
import types
from functools import partial

from neuroscribe.autodiff.grad import _align_gradient_shape, _build_graph
from neuroscribe.core._tensor_lib.backend.cpu.cpu_backend import CPUBackend

available_backends = {'cpu': CPUBackend}

try:
    from neuroscribe.core._tensor_lib.backend.cuda.cuda_backend import \
        CUDABackend
    available_backends['cuda'] = CUDABackend
except ImportError:
    pass

try:
    from neuroscribe.core._tensor_lib.backend.mps.mps_backend import MPSBackend
    available_backends['mps'] = MPSBackend
except ImportError:
    pass


class Tensor:

    _supported_backends = ['cpu', 'cuda', 'mps']
    _backends = available_backends

    def __init__(self, data, backend, requires_grad=False):
        self.data = data
        self.requires_grad = requires_grad
        self._backend = backend
        self._grad = None
        self._grad_fn = lambda: None
        self._prev = list()

    @property
    def device(self):
        return self._backend.device

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def shape(self):
        return self.data.shape

    @property
    def ndim(self):
        return self.data.ndim

    @property
    def strides(self):
        return self.data.strides

    @property
    def itemsize(self):
        return self.data.itemsize

    @property
    def is_cpu(self):
        return self.device == 'cpu'

    @property
    def is_cuda(self):
        return self.device == 'cuda'

    @property
    def is_mps(self):
        return self.device == 'mps'

    @property
    def grad(self):
        if self.requires_grad:
            if self._grad is None:
                self._grad = Tensor.zeros(self.shape, dtype=self.dtype, requires_grad=False, device=self.device)
            return self._grad
        else:
            return None

    @grad.setter
    def grad(self, value):
        self._grad = value

    @property
    def T(self):
        return self.transpose()

    def size(self, dim=None):
        return self.shape if dim is None else self.shape[dim]

    def numel(self):
        return math.prod(self.shape)

    def element_size(self):
        return self.itemsize

    def nbytes(self):
        return self.numel() * self.element_size()

    def normal_(self, mean, standard_deviation):
        self.data = self._backend.normal_(mean, standard_deviation, self.shape)

    def uniform_(self, lower_bound, upper_bound):
        self.data = self._backend.uniform_(lower_bound, upper_bound, self.shape)

    def zero_(self):
        self.data = self._backend.zeros(self.shape, dtype=self.dtype)

    def ones_(self):
        self.data = self._backend.ones(self.shape, dtype=self.dtype)

    def backward(self):
        if not self.requires_grad:
            raise RuntimeError('Gradient computation has not been enabled for this tensor.')

        graph = []
        visited = set()
        _build_graph(self, graph, visited)

        self.grad = Tensor.ones(self.shape, dtype=self.dtype, requires_grad=False, device=self.device)
        for tensor in reversed(graph):
            tensor._grad_fn()
        for tensor in reversed(graph):
            _align_gradient_shape(tensor)

    def __repr__(self):
        return f"Tensor({self.data}, dtype={self.dtype})"

    def __getitem__(self, index):
        if isinstance(index, slice) or isinstance(index, int):
            return Tensor.create(self.data[index], dtype=self.dtype, requires_grad=self.requires_grad, device=self.device)
        else:
            raise TypeError("Index must be an integer or slice object")

    def __setitem__(self, index, value):
        if isinstance(index, slice) or isinstance(index, int):
            self.data[index] = value
        else:
            raise TypeError("Index must be an integer or slice object")

    def __getstate__(self):
        state = self.__dict__.copy()
        if isinstance(state['_grad_fn'], types.LambdaType):
            state['_grad_fn'] = None
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        if self._grad_fn is None:
            self._grad_fn = lambda: None

    def __len__(self):
        return self.shape[0] if self.ndim > 0 else 1

    def __hash__(self):
        return id(self)

    def detach(self):
        if self.requires_grad == False:
            return self
        return Tensor(self.data, backend=self._backend, requires_grad=False)

    def item(self):
        if self.numel() != 1:
            raise ValueError(f"Tensor with {self.numel()} elements cannot be converted to Python scalar")
        return self.data.item()

    def argmax(self, dim=None):
        return Tensor(self._backend.argmax(self.data, dim), backend=self._backend, requires_grad=False)

    def argmin(self, dim=None):
        return Tensor(self._backend.argmin(self.data, dim), backend=self._backend, requires_grad=False)

    def is_contiguous(self):
        return self._backend.is_contiguous(self.data)

    def make_contiguous(self):
        if self.is_contiguous():
            return self
        return Tensor(self._backend.make_contiguous(self.data), backend=self._backend, requires_grad=self.requires_grad)

    def deep_copy(self):
        return Tensor(self._backend.deep_copy(self.data), backend=self._backend, requires_grad=self.requires_grad)

    def shallow_copy(self):
        return Tensor(self._backend.shallow_copy(self.data), backend=self._backend, requires_grad=self.requires_grad)

    @staticmethod
    def arange(start, stop=None, step=1, dtype='uint16', device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.arange(start, stop, step, dtype), backend=backend, requires_grad=False)

    @staticmethod
    def shuffle_(tensor):
        if not isinstance(tensor, Tensor):
            raise TypeError("The shuffle method expects a Tensor object.")
        if tensor.device == 'mps':
            tensor.data = tensor._backend.shuffle_(tensor.data)
        else:
            tensor._backend.shuffle_(tensor.data)

    @staticmethod
    def pad(input, pad, mode, *, constant_values):
        return Tensor(input._backend.pad(input.data, pad, mode, constant_values), backend=input._backend, requires_grad=input.requires_grad)

    # ********** Helper Methods **********
    @staticmethod
    def _get_backend(device):
        if device not in Tensor._supported_backends:
            raise ValueError(f"Unsupported device '{device}'. Supported devices are: {Tensor._supported_backends}.")
        if device not in Tensor._backends:
            raise ValueError(f"NeuroScribe is not installed with support for {device.upper()} devices.")
        return Tensor._backends[device]

    @staticmethod
    def _prepare_like_attributes(input, dtype, requires_grad, device):
        backend = Tensor._get_backend(device) if device is not None else input._backend
        dtype = dtype if dtype is not None else input.dtype
        requires_grad = requires_grad if requires_grad is not None else input.requires_grad
        return backend, dtype, requires_grad

    # ********** Creation Methods **********
    @staticmethod
    def create(data, dtype='float32', requires_grad=False, device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.create(data, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def zeros(shape, dtype='float32', requires_grad=False, device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.zeros(shape, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def zeros_like(input, dtype=None, requires_grad=None, device=None):
        backend, dtype, requires_grad = Tensor._prepare_like_attributes(input, dtype, requires_grad, device)
        return Tensor(backend.zeros_like(input.data, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def ones(shape, dtype='float32', requires_grad=False, device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.ones(shape, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def ones_like(input, dtype=None, requires_grad=None, device=None):
        backend, dtype, requires_grad = Tensor._prepare_like_attributes(input, dtype, requires_grad, device)
        return Tensor(backend.ones_like(input.data, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def randn(*shape, dtype='float32', requires_grad=False, device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.randn(*shape, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def rand(*shape, dtype='float32', requires_grad=False, device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.rand(*shape, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def randint(low, high=None, shape=None, *, dtype='int32', requires_grad=False, device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.randint(low, high, shape, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def empty(shape, dtype='float32', requires_grad=False, device='cpu'):
        backend = Tensor._get_backend(device)
        return Tensor(backend.empty(shape, dtype=dtype), backend=backend, requires_grad=requires_grad)

    @staticmethod
    def empty_like(input, dtype=None, requires_grad=None, device=None):
        backend, dtype, requires_grad = Tensor._prepare_like_attributes(input, dtype, requires_grad, device)
        return Tensor(backend.empty_like(input.data, dtype=dtype), backend=backend, requires_grad=requires_grad)

    def to(self, device):
        if self.device == device:
            return self
        return Tensor.create(self.data, dtype=self.dtype, requires_grad=self.requires_grad, device=device)

    def asnumpy(self):
        return self.to('cpu').data

    def astype(self, dtype):
        if self.dtype == dtype:
            return self
        return Tensor.create(self.data, dtype=dtype, requires_grad=self.requires_grad, device=self.device)

    # ********** Shape Manipulation Methods **********
    def flatten(self):
        return Tensor(self._backend.flatten(self.data), backend=self._backend, requires_grad=self.requires_grad)

    def flip(self, dims=None):
        return Tensor(self._backend.flip(self.data, dims), backend=self._backend, requires_grad=self.requires_grad)

    def squeeze(self, dim=None):
        return Tensor(self._backend.squeeze(self.data, dim), backend=self._backend, requires_grad=self.requires_grad)

    def reshape(self, shape):
        return Tensor(self._backend.reshape(self.data, shape), backend=self._backend, requires_grad=self.requires_grad)

    def transpose(self, dims=None):
        return Tensor(self._backend.transpose(self.data, dims), backend=self._backend, requires_grad=self.requires_grad)

    def split(self, indices_or_sections, dim=0):
        result = self._backend.split(self.data, indices_or_sections, dim)
        return [Tensor.create(t, dtype=self.dtype, requires_grad=self.requires_grad, device=self.device) for t in result]

    # ********** Tensor Operations **********
    def _exec_op(self, _op, *inputs, reverse=False):
        inputs = [Tensor.create(input, dtype=self.dtype, requires_grad=self.requires_grad, device=self.device)
                  if not isinstance(input, Tensor) else input for input in inputs]
        inputs = list(reversed([self] + inputs)) if reverse else [self] + inputs

        result_tensor = Tensor.create(_op(*inputs), dtype=self.dtype, requires_grad=False, device=self.device)

        if any(input.requires_grad for input in inputs):
            result_tensor.requires_grad = True
            result_tensor._grad_fn = partial(_op.backward, result_tensor)
            result_tensor._prev.extend(inputs)

        return result_tensor

    # ********** Reduction Ops **********
    def mean(self): return self._exec_op(self._backend.mean())
    def sum(self): return self._exec_op(self._backend.sum())
    def max(self): return self._exec_op(self._backend.max())
    def min(self): return self._exec_op(self._backend.min())

    # ********** Unary Ops **********
    def relu(self): return self._exec_op(self._backend.relu())
    def relu6(self): return self.relu() - (self - 6).relu()
    def leaky_relu(self, negative_slope=0.01): return self._exec_op(self._backend.leaky_relu(negative_slope))
    def elu(self, alpha=1.0): return self.relu() - alpha * (1 - self.exp()).relu()
    def swish(self): return self * self.sigmoid()
    def silu(self): return self.swish()
    def sigmoid(self): return self._exec_op(self._backend.sigmoid())
    def softmax(self): return (self - self.max()).exp() / (self - self.max()).exp().sum()
    def log_softmax(self): return (self - self.max()) - (self - self.max()).exp().sum().log()
    def softmin(self): return ((self - self.max()).neg()).exp() / ((self - self.max()).neg()).exp().sum()
    def square(self): return self._exec_op(self._backend.square())
    def neg(self): return self._exec_op(self._backend.neg())
    def clip(self, min, max): return self._exec_op(self._backend.clip(min, max))
    def sign(self): return self._exec_op(self._backend.sign())
    def abs(self): return self * self.sign()
    def reciprocal(self): return self._exec_op(self._backend.reciprocal())
    def sqrt(self): return self._exec_op(self._backend.sqrt())
    def rsqrt(self): return self.reciprocal().sqrt()
    def log(self): return self._exec_op(self._backend.log())
    def log10(self): return self.log() / math.log(10)
    def log2(self): return self.log() / math.log(2)
    def log1p(self): return (self + 1).log()
    def exp(self): return self._exec_op(self._backend.exp())
    def exp2(self): return (self * math.log(2)).exp()
    def sin(self): return self._exec_op(self._backend.sin())
    def cos(self): return self._exec_op(self._backend.cos())
    def tan(self): return self.sin() / self.cos()
    # NOTE: 2.0 * ((2.0 * self).sigmoid()) - 1.0 could be used instead, but it may result in lower precision.
    def tanh(self): return self._exec_op(self._backend.tanh())
    def sinh(self): return (self.exp() - self.neg().exp()) / 2
    def cosh(self): return (self.exp() + self.neg().exp()) / 2
    def atanh(self): return ((1 + self) / (1 - self)).log() / 2
    def asinh(self): return (self + (self.square() + 1).sqrt()).log()
    def acosh(self): return (self + (self.square() - 1).sqrt()).log()
    def hardtanh(self, min=-1, max=1): return self.clip(min, max)
    def gelu(self): return 0.5 * self * (1 + (math.sqrt(2 / math.pi) * (self + 0.044715 * self ** 3)).tanh())
    def softplus(self, beta=1.0): return (1 / beta) * (1 + (self * beta).exp()).log()
    def mish(self): return self * self.softplus().tanh()
    def softsign(self): return self / (1 + self.abs())

    # ********** Binary Ops **********
    def pow(self, exponent, reverse=False): return self._exec_op(self._backend.pow(), exponent, reverse=reverse)
    def add(self, other, reverse=False): return self._exec_op(self._backend.add(), other, reverse=reverse)
    def sub(self, other, reverse=False): return self._exec_op(self._backend.sub(), other, reverse=reverse)
    def mul(self, other, reverse=False): return self._exec_op(self._backend.mul(), other, reverse=reverse)
    def div(self, other, reverse=False): return self._exec_op(self._backend.div(), other, reverse=reverse)
    def matmul(self, other, reverse=False): return self._exec_op(self._backend.matmul(), other, reverse=reverse)

    def __pow__(self, other): return self.pow(other)
    def __add__(self, other): return self.add(other)
    def __sub__(self, other): return self.sub(other)
    def __mul__(self, other): return self.mul(other)
    def __truediv__(self, other): return self.div(other)
    def __matmul__(self, other): return self.matmul(other)

    def __rpow__(self, other): return self.pow(other, reverse=True)
    def __radd__(self, other): return self.add(other, reverse=True)
    def __rsub__(self, other): return self.sub(other, reverse=True)
    def __rmul__(self, other): return self.mul(other, reverse=True)
    def __rtruediv__(self, other): return self.div(other, reverse=True)
    def __rmatmul__(self, other): return self.matmul(other, reverse=True)



from .activation import (ELU, GELU, HardTanh, LeakyReLU, LogSoftmax, Mish,
                         ReLU, ReLU6, Sigmoid, SiLU, Softmax, Softmin,
                         Softplus, Softsign, Swish, Tanh)
from .container import Sequential
from .conv import Conv1d, Conv2d, Conv3d
from .linear import Linear
from .loss import BCELoss, L1Loss, MSELoss
from .module import Module

__all__ = ['Linear', 'Module', 'ReLU', 'ReLU6', 'LeakyReLU', 'ELU', 'Swish', 'SiLU', 'Tanh', 'HardTanh', 'GELU', 'Sigmoid', 'Softmax', 'LogSoftmax', 'Softmin',
           'Softplus', 'Mish', 'Softsign', 'MSELoss', 'L1Loss', 'BCELoss', 'Conv1d', 'Conv2d', 'Conv3d', 'Sequential']



from collections.abc import Iterable
from itertools import repeat

import numpy as np


def _ntuple(n, name="parse"):
    def parse(x):
        if isinstance(x, Iterable):
            return tuple(x)
        return tuple(repeat(x, n))

    parse.__name__ = name
    return parse


_single = _ntuple(1, "_single")
_pair = _ntuple(2, "_pair")
_triple = _ntuple(3, "_triple")


def _calculate_padding(padding, kernel_size, dim):
    if dim not in [1, 2, 3]:
        raise ValueError(f"Invalid dim value '{dim}'. Expected one of [1, 2, 3]")

    _ntuple_fn = {1: _single, 2: _pair, 3: _triple}
    kernel_size = _ntuple_fn[dim](kernel_size)

    if isinstance(padding, str):
        if padding == 'same':
            pad = []
            for k in kernel_size:
                if k % 2 == 0:
                    pad.append((k // 2 - 1, k // 2))
                else:
                    pad.append((k // 2, k // 2))
            return tuple(pad)
        elif padding == 'valid':
            return tuple((0, 0) for _ in kernel_size)
        else:
            raise ValueError(f"Invalid string padding value '{padding}'. Expected either 'same', 'valid'")
    elif isinstance(padding, tuple):
        pad = [_pair(p) for p in padding] if dim > 1 else [_pair(padding)]
        return tuple(pad)
    elif isinstance(padding, int):
        return tuple((_pair(padding)) for _ in kernel_size)
    else:
        raise ValueError(f"Invalid padding value '{padding}'. Expected either 'same', 'valid', a tuple, or an integer")


def _sliding_window_view(input, window_shape, step_shape, conv_dim):

    if conv_dim not in [1, 2, 3]:
        raise ValueError(f"Invalid conv_dim value '{conv_dim}'. Expected one of [1, 2, 3]")

    if conv_dim == 1:
        # For 1D convolution
        new_shape = (
            input.shape[0],                                           # batch size
            input.shape[1],                                           # channels
            (input.shape[2] - window_shape[0]) // step_shape[0] + 1,  # new length
            window_shape[0]                                           # window length
        )
        new_strides = (
            input.strides[0],                   # stride to move to the next batch
            input.strides[1],                   # stride to move to the next channel
            input.strides[2] * step_shape[0],   # stride to move to the next window along the length
            input.strides[2]                    # stride to move to the next element within the window along the length
        )
    elif conv_dim == 2:
        # For 2D convolution
        new_shape = (
            input.shape[0],                                           # batch size
            input.shape[1],                                           # number of input channels
            (input.shape[2] - window_shape[0]) // step_shape[0] + 1,  # number of windows along height
            (input.shape[3] - window_shape[1]) // step_shape[1] + 1,  # number of windows along width
            window_shape[0],                                          # height of each window (kernel height)
            window_shape[1]                                           # width of each window (kernel width)
        )
        new_strides = (
            input.strides[0],                   # stride to move to the next batch
            input.strides[1],                   # stride to move to the next channel
            input.strides[2] * step_shape[0],   # stride to move to the next window along height
            input.strides[3] * step_shape[1],   # stride to move to the next window along width
            input.strides[2],                   # stride to move to the next element within the window along height
            input.strides[3]                    # stride to move to the next element within the window along width
        )
    elif conv_dim == 3:
        # For 3D convolution
        new_shape = (
            input.shape[0],                                           # batch size
            input.shape[1],                                           # channels
            (input.shape[2] - window_shape[0]) // step_shape[0] + 1,  # number of windows along depth
            (input.shape[3] - window_shape[1]) // step_shape[1] + 1,  # number of windows along height
            (input.shape[4] - window_shape[2]) // step_shape[2] + 1,  # number of windows along width
            window_shape[0],                                          # depth of each window (kernel depth)
            window_shape[1],                                          # height of each window (kernel height)
            window_shape[2]                                           # width of each window (kernel width)
        )
        new_strides = (
            input.strides[0],                   # stride to move to the next batch
            input.strides[1],                   # stride to move to the next channel
            input.strides[2] * step_shape[0],   # stride to move to the next window along depth
            input.strides[3] * step_shape[1],   # stride to move to the next window along height
            input.strides[4] * step_shape[2],   # stride to move to the next window along width
            input.strides[2],                   # stride to move to the next element within the window along depth
            input.strides[3],                   # stride to move to the next element within the window along height
            input.strides[4]                    # stride to move to the next element within the window along width
        )

    return np.lib.stride_tricks.as_strided(input.data, shape=new_shape, strides=new_strides, writeable=False)



from neuroscribe.core._tensor_lib._tensor import Tensor

from .optimizer import Optimizer


class Adam(Optimizer):
    def __init__(self, params, lr=0.001, betas=(0.9, 0.999), eps=1e-8):
        super().__init__(params)
        if lr < 0:
            raise ValueError(f"Invalid learning rate: {lr}")
        self.lr = lr
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        self.beta1, self.beta2 = betas
        if eps < 0:
            raise ValueError(f"Invalid epsilon value: {eps}")
        self.eps = eps
        self.m = [Tensor.zeros_like(param) for param in self.params]
        self.v = [Tensor.zeros_like(param) for param in self.params]
        self.t = 0

    def step(self):
        self.t += 1
        for i, param in enumerate(self.params):
            self.m[i].data = self.beta1 * self.m[i].data + (1 - self.beta1) * param.grad.data
            self.v[i].data = self.beta2 * self.v[i].data + (1 - self.beta2) * (param.grad.data ** 2)
            m_hat = self.m[i].data / (1 - self.beta1 ** self.t)
            v_hat = self.v[i].data / (1 - self.beta2 ** self.t)
            param.data = param.data - self.lr * m_hat / (v_hat ** 0.5 + self.eps)



from ._functional_api import *
from ._tensor import Tensor
from .einsumfunc import einsum

__all__ = ['Tensor', 'einsum', 'tensor', 'arange', 'shuffle_', 'zeros', 'zeros_like', 'ones', 'ones_like', 'randn', 'rand', 'randint', 'empty', 'empty_like', 'argmax',
           'argmin', 'add', 'sub', 'mul', 'div', 'matmul', 'relu', 'relu6', 'leaky_relu', 'tanh', 'sinh', 'cosh', 'hardtanh', 'sigmoid', 'softmax', 'log_softmax',
           'softmin', 'mean', 'sum', 'max', 'min', 'square', 'neg', 'clip', 'sign', 'abs', 'reciprocal', 'sqrt', 'rsqrt', 'log', 'log10', 'log2', 'log1p',
           'exp', 'exp2', 'sin', 'cos', 'tan', 'atanh', 'asinh', 'acosh', 'gelu', 'elu', 'swish', 'silu', 'softplus', 'mish', 'softsign', 'pow', 'pad',
           'flatten', 'flip', 'squeeze', 'reshape', 'transpose', 'split']



import csv
import gzip
import os
import shutil
from struct import calcsize, unpack

import numpy as np
import requests
from tqdm import tqdm

from neuroscribe.core._tensor_lib._tensor import Tensor


def _request_file(url):
    response = requests.get(url, stream=True)
    response.raise_for_status()
    return response


def _save_file(response, file_path):
    total_size = int(response.headers.get('content-length', 0))
    with open(file_path, 'wb') as f, tqdm(
        desc=f"Downloading {os.path.basename(file_path)}",
        total=total_size,
        unit='B',
        unit_scale=True,
        unit_divisor=1024,
    ) as bar:
        for data in response.iter_content(chunk_size=1024):
            f.write(data)
            bar.update(len(data))


def _decompress_file(file_path):
    if file_path.endswith('.gz'):
        decompressed_file_path = file_path[:-3]
        if not os.path.exists(decompressed_file_path):
            with gzip.open(file_path, 'rb') as f_in:
                with open(decompressed_file_path, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
            os.remove(file_path)  # Remove the .gz file


def _read_delimited_data(file_path, chunk_size, delimiter):
    data = []
    with open(file_path, 'r') as f:
        reader = csv.reader(f, delimiter=delimiter)
        chunk = []
        for row in reader:
            chunk.append([int(x) for x in row])
            if len(chunk) >= chunk_size:
                data.append(np.array(chunk, dtype=np.uint8))
                chunk = []
        if chunk:
            data.append(np.array(chunk, dtype=np.uint8))
    return data


def _read_ubyte_data(file_path, chunk_size):
    data = []
    with open(file_path, 'rb') as f:
        magic, num, rows, cols = _read_binary_header(f, '>4I')
        for _ in range(num):
            img = np.fromfile(f, dtype=np.uint8, count=rows * cols)
            img = img.reshape((rows, cols))
            data.append(img)
    return data


def _read_ubyte_label_data(file_path, chunk_size):
    data = []
    with open(file_path, 'rb') as f:
        magic, num = _read_binary_header(f, '>2I')
        for _ in range(num):
            label = np.fromfile(f, dtype=np.uint8, count=1)
            data.append(label)
    return data


def _read_binary_header(file, format_string):
    header_size = calcsize(format_string)
    header_bytes = file.read(header_size)
    return unpack(format_string, header_bytes)


def _convert_to_numpy(*inputs):
    validated_inputs = []
    for input in inputs:
        if isinstance(input, Tensor):
            validated_inputs.append(input.asnumpy())
        elif isinstance(input, np.ndarray):
            validated_inputs.append(input)
        else:
            validated_inputs.append(np.array(input))
    return validated_inputs


def _ensure_ndim(input, ndim):
    (input,) = _convert_to_numpy(input)
    print(input)
    if input.ndim != ndim:
        raise ValueError(f"Input data must be {ndim}-dimensional")
    return input



import neuroscribe as ns
from neuroscribe.autodiff.function import Function
from neuroscribe.core._tensor_lib._tensor import Tensor

from .utils import (_calculate_padding, _pair, _single, _sliding_window_view,
                    _triple)


class Conv1d(Function):
    def __init__(self, weight, bias, stride, padding, dilation, groups):
        self.weight = weight
        self.bias = bias
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups

    def forward(self, input):
        batch_size = input.shape[0]
        out_channels, _, kernel_size = self.weight.shape
        stride = _single(self.stride)

        (pad,) = _calculate_padding(self.padding, kernel_size=kernel_size, dim=1)

        input_padded = ns.pad(
            input, ((0, 0), (0, 0), pad), mode='constant', constant_values=0
        )
        windows = Tensor(
            _sliding_window_view(input_padded.asnumpy(), window_shape=(kernel_size,), step_shape=stride, conv_dim=1),
            backend=input._backend,
            requires_grad=input.requires_grad
        )

        windows_reshaped = windows.transpose((0, 2, 1, 3)).reshape((batch_size, windows.shape[2], -1))
        weight_reshaped = self.weight.reshape((out_channels, -1))

        output = ns.einsum("bwc,kc->bwk", windows_reshaped, weight_reshaped).transpose((0, 2, 1))

        if self.bias is not None:
            output.data += self.bias.data[:, None]

        return output.data

    def backward(self, result_tensor):
        (input,) = result_tensor._prev
        kernel_size = self.weight.shape[2]

        # Input gradient
        weight_flipped = ns.flip(self.weight, dims=-1)
        if self.padding in ['valid', 'same']:
            (output_grad_pad,) = _calculate_padding(padding='same', kernel_size=kernel_size, dim=1)
        else:
            (output_grad_pad,) = _calculate_padding(padding=self.padding, kernel_size=kernel_size, dim=1)
        output_grad_pad = tuple(reversed(output_grad_pad)) if self.padding == 'same' else output_grad_pad
        output_grad_padded = ns.pad(result_tensor.grad, ((0, 0), (0, 0), output_grad_pad), mode='constant', constant_values='0')
        if self.padding == 'valid':
            input_grad = Conv1d(
                weight_flipped.transpose((1, 0, 2)),
                bias=None,
                stride=self.stride,
                padding='same',
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)
        elif self.padding == 'same':
            input_grad = Conv1d(
                weight_flipped.transpose((1, 0, 2)),
                bias=None,
                stride=self.stride,
                padding='valid',
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)
        else:
            input_grad = Conv1d(
                weight_flipped.transpose((1, 0, 2)),
                bias=None,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)

        # Crop input_grad to match input shape
        if input_grad.shape[2] != input.shape[2]:
            diff = input_grad.shape[2] - input.shape[2]
            input_grad = input_grad[:, :, diff // 2:-(diff - diff // 2)]

        # Weight gradient
        (pad,) = _calculate_padding(self.padding, kernel_size=kernel_size, dim=1)
        input_padded = ns.pad(input, ((0, 0), (0, 0), pad), mode='constant', constant_values='0')
        weight_grad = Conv1d(
            result_tensor.grad.transpose((1, 0, 2)),
            bias=None,
            stride=self.stride,
            padding='valid',
            dilation=self.dilation,
            groups=self.groups
        )(input_padded.transpose((1, 0, 2))).transpose((1, 0, 2))

        input.grad.data = input.grad.data + input_grad
        self.weight.grad.data = self.weight.grad.data + weight_grad
        if self.bias:
            self.bias.grad.data = self.bias.grad.data + result_tensor.grad.data.sum(axis=(0, 2))


class Conv2d(Function):
    def __init__(self, weight, bias, stride, padding, dilation, groups):
        self.weight = weight
        self.bias = bias
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups

    def forward(self, input):
        batch_size = input.shape[0]
        out_channels, _, kernel_height, kernel_width = self.weight.shape
        stride_height, stride_width = _pair(self.stride)

        pad_height, pad_width = _calculate_padding(self.padding, kernel_size=(kernel_height, kernel_width), dim=2)

        input_padded = ns.pad(
            input, ((0, 0), (0, 0), pad_height, pad_width), mode='constant', constant_values=0
        )
        windows = Tensor(
            _sliding_window_view(
                input_padded.asnumpy(),
                window_shape=(kernel_height, kernel_width),
                step_shape=(stride_height, stride_width),
                conv_dim=2
            ),
            backend=input._backend,
            requires_grad=input.requires_grad
        )

        windows_reshaped = windows.transpose((0, 2, 3, 1, 4, 5)).reshape(
            (batch_size, windows.shape[2], windows.shape[3], -1)
        )
        weight_reshaped = self.weight.reshape((out_channels, -1))

        output = ns.einsum("bhwc,kc->bhwk", windows_reshaped, weight_reshaped).transpose((0, 3, 1, 2))

        if self.bias is not None:
            output.data += self.bias.data[:, None, None]

        return output.data

    def backward(self, result_tensor):
        (input,) = result_tensor._prev
        kernel_height, kernel_width = self.weight.shape[2:]

        # Input gradient
        weight_flipped = ns.flip(ns.flip(self.weight, dims=-1), dims=-2)
        if self.padding in ['valid', 'same']:
            output_grad_pad = _calculate_padding(padding='same', kernel_size=(kernel_height, kernel_width), dim=2)
        else:
            output_grad_pad = _calculate_padding(padding=self.padding, kernel_size=(kernel_height, kernel_width), dim=2)
        output_grad_pad = tuple(reversed(output_grad_pad)) if self.padding == 'same' else output_grad_pad
        output_grad_padded = ns.pad(
            result_tensor.grad, ((0, 0), (0, 0), output_grad_pad[0], output_grad_pad[1]), mode='constant', constant_values=0
        )
        if self.padding == 'valid':
            input_grad = Conv2d(
                weight_flipped.transpose((1, 0, 2, 3)),
                bias=None,
                stride=self.stride,
                padding='same',
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)
        elif self.padding == 'same':
            input_grad = Conv2d(
                weight_flipped.transpose((1, 0, 2, 3)),
                bias=None,
                stride=self.stride,
                padding='valid',
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)
        else:
            input_grad = Conv2d(
                weight_flipped.transpose((1, 0, 2, 3)),
                bias=None,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)

        # Crop input_grad to match input shape
        if input_grad.shape[2:] != input.shape[2:]:
            diff_h = input_grad.shape[2] - input.shape[2]
            diff_w = input_grad.shape[3] - input.shape[3]
            input_grad = input_grad[:, :, diff_h // 2:-(diff_h - diff_h // 2), diff_w // 2:-(diff_w - diff_w // 2)]

        # Weight gradient
        pad_height, pad_width = _calculate_padding(self.padding, kernel_size=(kernel_height, kernel_width), dim=2)
        input_padded = ns.pad(input, ((0, 0), (0, 0), pad_height, pad_width), mode='constant', constant_values=0)
        weight_grad = Conv2d(
            result_tensor.grad.transpose((1, 0, 2, 3)),
            bias=None,
            stride=self.stride,
            padding='valid',
            dilation=self.dilation,
            groups=self.groups
        )(input_padded.transpose((1, 0, 2, 3))).transpose((1, 0, 2, 3))

        input.grad.data = input.grad.data + input_grad
        self.weight.grad.data = self.weight.grad.data + weight_grad
        if self.bias:
            self.bias.grad.data = self.bias.grad.data + result_tensor.grad.data.sum(axis=(0, 2, 3))


class Conv3d(Function):
    def __init__(self, weight, bias, stride, padding, dilation, groups):
        self.weight = weight
        self.bias = bias
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups

    def forward(self, input):
        batch_size = input.shape[0]
        out_channels, _, kernel_depth, kernel_height, kernel_width = self.weight.shape
        stride_depth, stride_height, stride_width = _triple(self.stride)

        pad_depth, pad_height, pad_width = _calculate_padding(
            self.padding, kernel_size=(kernel_depth, kernel_height, kernel_width), dim=3
        )

        input_padded = ns.pad(
            input, ((0, 0), (0, 0), pad_depth, pad_height, pad_width), mode='constant', constant_values=0
        )
        windows = Tensor(
            _sliding_window_view(
                input_padded.asnumpy(),
                window_shape=(kernel_depth, kernel_height, kernel_width),
                step_shape=(stride_depth, stride_height, stride_width),
                conv_dim=3
            ),
            backend=input._backend,
            requires_grad=input.requires_grad
        )

        windows_reshaped = windows.transpose((0, 2, 3, 4, 1, 5, 6, 7)).reshape(
            (batch_size, windows.shape[2], windows.shape[3], windows.shape[4], -1)
        )
        weight_reshaped = self.weight.reshape((out_channels, -1))

        output = ns.einsum("bdhwc,kc->bdhwk", windows_reshaped, weight_reshaped).transpose((0, 4, 1, 2, 3))

        if self.bias is not None:
            output.data += self.bias.data[:, None, None, None]

        return output.data

    def backward(self, result_tensor):
        (input,) = result_tensor._prev
        kernel_depth, kernel_height, kernel_width = self.weight.shape[2:]

        # Input gradient
        weight_flipped = ns.flip(ns.flip(ns.flip(self.weight, dims=-1), dims=-2), dims=-3)
        if self.padding in ['valid', 'same']:
            output_grad_pad = _calculate_padding(
                padding='same', kernel_size=(kernel_depth, kernel_height, kernel_width), dim=3
            )
        else:
            output_grad_pad = _calculate_padding(
                padding=self.padding, kernel_size=(kernel_depth, kernel_height, kernel_width), dim=3
            )
        output_grad_pad = tuple(reversed(output_grad_pad)) if self.padding == 'same' else output_grad_pad
        output_grad_padded = ns.pad(
            result_tensor.grad, ((0, 0), (0, 0), output_grad_pad[0], output_grad_pad[1], output_grad_pad[2]), mode='constant', constant_values=0
        )
        if self.padding == 'valid':
            input_grad = Conv3d(
                weight_flipped.transpose((1, 0, 2, 3, 4)),
                bias=None,
                stride=self.stride,
                padding='same',
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)
        elif self.padding == 'same':
            input_grad = Conv3d(
                weight_flipped.transpose((1, 0, 2, 3, 4)),
                bias=None,
                stride=self.stride,
                padding='valid',
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)
        else:
            input_grad = Conv3d(
                weight_flipped.transpose((1, 0, 2, 3, 4)),
                bias=None,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
                groups=self.groups
            )(output_grad_padded)

        # Crop input_grad to match input shape
        if input_grad.shape[2:] != input.shape[2:]:
            diff_d = input_grad.shape[2] - input.shape[2]
            diff_h = input_grad.shape[3] - input.shape[3]
            diff_w = input_grad.shape[4] - input.shape[4]
            input_grad = input_grad[:, :,
                                    diff_d // 2:-(diff_d - diff_d // 2),
                                    diff_h // 2:-(diff_h - diff_h // 2),
                                    diff_w // 2:-(diff_w - diff_w // 2)]

        # Weight gradient
        pad_depth, pad_height, pad_width = _calculate_padding(
            self.padding, kernel_size=(kernel_depth, kernel_height, kernel_width), dim=3
        )
        input_padded = ns.pad(input, ((0, 0), (0, 0), pad_depth, pad_height, pad_width), mode='constant', constant_values=0)
        weight_grad = Conv3d(
            result_tensor.grad.transpose((1, 0, 2, 3, 4)),
            bias=None,
            stride=self.stride,
            padding='valid',
            dilation=self.dilation,
            groups=self.groups
        )(input_padded.transpose((1, 0, 2, 3, 4))).transpose((1, 0, 2, 3, 4))

        input.grad.data = input.grad.data + input_grad
        self.weight.grad.data = self.weight.grad.data + weight_grad
        if self.bias:
            self.bias.grad.data = self.bias.grad.data + result_tensor.grad.data.sum(axis=(0, 2, 3, 4))



import neuroscribe.optim as optim
from neuroscribe.nn.modules.activation import (ELU, GELU, HardTanh, LeakyReLU,
                                               LogSoftmax, Mish, ReLU, ReLU6,
                                               Sigmoid, SiLU, Softmax, Softmin,
                                               Softplus, Softsign, Swish, Tanh)
from neuroscribe.nn.modules.conv import Conv1d, Conv2d, Conv3d
from neuroscribe.nn.modules.loss import BCELoss, L1Loss, MSELoss
from neuroscribe.utils.metrics import (accuracy_score, confusion_matrix,
                                       f1_score, mean_absolute_error,
                                       mean_squared_error, precision_score,
                                       r_squared, recall_score,
                                       root_mean_squared_error)

# ********** Simplified NN Mappings **********
activation_function_map = {
    'relu': ReLU(),
    'relu6': ReLU6(),
    'leaky_relu': LeakyReLU(),
    'elu': ELU(),
    'swish': Swish(),
    'silu': SiLU(),
    'tanh': Tanh(),
    'hardtanh': HardTanh(),
    'gelu': GELU(),
    'sigmoid': Sigmoid(),
    'softmax': Softmax(),
    'log_softmax': LogSoftmax(),
    'softmin': Softmin(),
    'softplus': Softplus(),
    'mish': Mish(),
    'softsign': Softsign(),
}

loss_function_map = {
    "binary_cross_entropy": BCELoss,
    "mean_absolute_error": L1Loss,
    "mean_squared_error": MSELoss
}

optimizer_map = {
    'adam': optim.Adam,
    'sgd': optim.SGD,
    'sgd_momentum': optim.SGD,
}

metric_map = {
    'precision': precision_score,
    'recall': recall_score,
    'f1': f1_score,
    'accuracy': accuracy_score,
    'confusion_matrix': confusion_matrix,
    'r_squared': r_squared,
    'mean_squared_error': mean_squared_error,
    'mean_absolute_error': mean_absolute_error,
    'root_mean_squared_error': root_mean_squared_error,
}

conv_layer_map = {
    1: Conv1d,
    2: Conv2d,
    3: Conv3d
}



import os
import platform
from pathlib import Path

from setuptools import find_packages, setup

BASE_DIR = Path(__file__).parent


def cuda_available():
    return 'CUDA_PATH' in os.environ or 'CUDA_HOME' in os.environ


def on_mac_silicon():
    return platform.system() == "Darwin" and "arm" in platform.machine().lower()


install_requires = ['numpy==1.26.4', 'tqdm==4.65.0', 'requests==2.31.0', 'plotly==5.22.0']
if cuda_available():
    install_requires.extend(['cupy-cuda12x==13.0.0', 'fastrlock==0.8.2'])
if on_mac_silicon():
    install_requires.extend(['jax-metal==0.1.0', 'jaxlib==0.4.30', 'jax==0.4.30'])
long_description = (BASE_DIR / "README.md").read_text()


setup(
    name='neuro-scribe',
    version='0.2.0.post3',
    author='Ghaith Husrieh',
    description='NeuroScribe - a lightweight deep learning framework.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    project_urls={
        'Source': 'https://github.com/Ghaith-Husrieh/NeuroScribe'
    },
    classifiers=[
        'Programming Language :: Python :: 3.11',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.11',
    install_requires=install_requires
)



import numpy as np

from neuroscribe.autodiff.function import Function


# ********** Unary ops **********
class ReLU(Function):
    def forward(self, t1): return np.maximum(0, t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + (result_tensor.data > 0) * result_tensor.grad.data


class LeakyReLU(Function):
    def __init__(self, negative_slope):
        self.negative_slope = negative_slope

    def forward(self, t1):
        return np.maximum(self.negative_slope * t1.data, t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + np.where(t1.data > 0, 1, self.negative_slope) * result_tensor.grad.data


class Tanh(Function):
    def forward(self, t1):
        return np.tanh(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + (1 - np.tanh(t1.data)**2) * result_tensor.grad.data


class Sigmoid(Function):
    def forward(self, t1):
        return 1 / (1 + np.exp(-t1.data))

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        sigmoid = 1 / (1 + np.exp(-t1.data))
        t1.grad.data = t1.grad.data + sigmoid * (1 - sigmoid) * result_tensor.grad.data


class Mean(Function):
    def forward(self, t1): return np.mean(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + result_tensor.grad.data / t1.data.size


class Sum(Function):
    def forward(self, t1): return np.sum(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + 1 * result_tensor.grad.data


class Max(Function):
    def forward(self, t1): return np.max(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1_partial_grad = np.zeros_like(t1.data)
        max_positions = np.where(t1.data == result_tensor.data)[0]
        t1_partial_grad[max_positions] = 1.0 / len(max_positions)
        t1.grad.data = t1.grad.data + t1_partial_grad * result_tensor.grad.data


class Min(Function):
    def forward(self, t1): return np.min(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1_partial_grad = np.zeros_like(t1.data)
        min_positions = np.where(t1.data == result_tensor.data)[0]
        t1_partial_grad[min_positions] = 1.0 / len(min_positions)
        t1.grad.data = t1.grad.data + t1_partial_grad * result_tensor.grad.data


class Square(Function):
    def forward(self, t1): return np.square(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + 2 * t1.data * result_tensor.grad.data


class Neg(Function):
    def forward(self, t1): return np.negative(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + (-1 * result_tensor.grad.data)


class Clip(Function):
    def __init__(self, min, max):
        self.min = min
        self.max = max

    def forward(self, t1): return np.clip(t1.data, self.min, self.max)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + np.where(
            (t1.data <= self.min) | (t1.data >= self.max),
            0,
            result_tensor.grad.data
        )


class Sign(Function):
    def forward(self, t1):
        return np.sign(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + 0 * result_tensor.grad.data


class Reciprocal(Function):
    def forward(self, t1): return np.reciprocal(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + (-1 / (t1.data**2)) * result_tensor.grad.data


class Sqrt(Function):
    def forward(self, t1): return np.sqrt(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + (1 / (2 * np.sqrt(t1.data))) * result_tensor.grad.data


class Log(Function):
    def forward(self, t1): return np.log(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + (1 / t1.data) * result_tensor.grad.data


class Exp(Function):
    def forward(self, t1): return np.exp(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + np.exp(t1.data) * result_tensor.grad.data


class Sin(Function):
    def forward(self, t1): return np.sin(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + np.cos(t1.data) * result_tensor.grad.data


class Cos(Function):
    def forward(self, t1): return np.cos(t1.data)

    def backward(self, result_tensor):
        (t1,) = result_tensor._prev
        t1.grad.data = t1.grad.data + -np.sin(t1.data) * result_tensor.grad.data


# ********** Binary ops **********
class Pow(Function):
    def forward(self, t1, t2): return np.power(t1.data, t2.data)

    def backward(self, result_tensor):
        t1, t2 = result_tensor._prev
        t1.grad.data = t1.grad.data + t2.data * np.power(t1.data, t2.data - 1) * result_tensor.grad.data
        t2.grad.data = t2.grad.data + result_tensor.data * np.log(t1.data) * result_tensor.grad.data


class Add(Function):
    def forward(self, t1, t2): return t1.data + t2.data

    def backward(self, result_tensor):
        t1, t2 = result_tensor._prev
        t1.grad.data = t1.grad.data + result_tensor.grad.data
        t2.grad.data = t2.grad.data + result_tensor.grad.data


class Sub(Function):
    def forward(self, t1, t2): return t1.data - t2.data

    def backward(self, result_tensor):
        t1, t2 = result_tensor._prev
        t1.grad.data = t1.grad.data + result_tensor.grad.data
        t2.grad.data = t2.grad.data - result_tensor.grad.data


class Mul(Function):
    def forward(self, t1, t2): return t1.data * t2.data

    def backward(self, result_tensor):
        t1, t2 = result_tensor._prev
        t1.grad.data = t1.grad.data + t2.data * result_tensor.grad.data
        t2.grad.data = t2.grad.data + t1.data * result_tensor.grad.data


class Div(Function):
    def forward(self, t1, t2): return t1.data / t2.data

    def backward(self, result_tensor):
        t1, t2 = result_tensor._prev
        t1.grad.data = t1.grad.data + (1 / t2.data) * result_tensor.grad.data
        t2.grad.data = t2.grad.data + (-t1.data / (t2.data**2)) * result_tensor.grad.data


class MatMul(Function):
    def forward(self, t1, t2): return np.matmul(t1.data, t2.data)

    def backward(self, result_tensor):
        t1, t2 = result_tensor._prev
        t1.grad.data = t1.grad.data + np.matmul(result_tensor.grad.data, t2.data.T)
        t2.grad.data = t2.grad.data + np.matmul(t1.data.T, result_tensor.grad.data)



import neuroscribe.nn.functional as F

from .module import Module


class MSELoss(Module):
    def __init__(self):
        super().__init__()

    def forward(self, predictions, targets):
        return F.mse_loss(predictions, targets)


class L1Loss(Module):
    def __init__(self):
        super().__init__()

    def forward(self, predictions, targets):
        return F.l1_loss(predictions, targets)


class BCELoss(Module):
    def __init__(self):
        super().__init__()

    def forward(self, predictions, targets):
        return F.binary_cross_entropy(predictions, targets)



import math

import neuroscribe.nn.functional as F
from neuroscribe.core._nn.utils import _pair, _single, _triple
from neuroscribe.core._tensor_lib._tensor import Tensor

from .. import init
from .module import Module


class _ConvNd(Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride, padding, dilation, transposed, output_padding, groups, padding_mode, bias=True):
        super().__init__()
        if groups <= 0:
            raise ValueError('groups must be a positive integer')
        if in_channels % groups != 0:
            raise ValueError('in_channels must be divisible by groups')
        if out_channels % groups != 0:
            raise ValueError('out_channels must be divisible by groups')
        valid_padding_strings = {'same', 'valid'}
        if isinstance(padding, str):
            if padding not in valid_padding_strings:
                raise ValueError(
                    f"Invalid padding string {padding!r}, should be one of {valid_padding_strings}")
            if padding == 'same' and any(s != 1 for s in stride):
                raise ValueError("padding='same' is not supported for strided convolutions")
        valid_padding_modes = {'zeros', 'reflect', 'replicate', 'circular'}
        if padding_mode not in valid_padding_modes:
            raise ValueError(
                f"padding_mode must be one of {valid_padding_modes}, but got padding_mode='{padding_mode}'"
            )
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.output_padding = output_padding
        self.groups = groups
        self.padding_mode = padding_mode
        if transposed:
            self.weight = Tensor.empty((in_channels, out_channels // groups, *kernel_size))
        else:
            self.weight = Tensor.empty((out_channels, in_channels // groups, *kernel_size))
        self.bias = Tensor.empty(out_channels) if bias else None
        self.init_parameters()

    def init_parameters(self):
        init.kaiming_uniform(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in = init.calculate_correct_fan(self.weight, mode='fan_in')
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            self.bias.uniform_(-bound, bound)


class Conv1d(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, padding_mode='zeros', bias=True):
        kernel_size = _single(kernel_size)
        stride = _single(stride)
        padding = padding if isinstance(padding, str) else padding
        dilation = _single(dilation)
        super().__init__(in_channels, out_channels, kernel_size, stride, padding,
                         dilation, False, _single(0), groups, padding_mode, bias)

    def _conv_forward(self, input, weight, bias):
        if self.padding_mode != 'zeros':
            raise NotImplementedError("Padding modes other than 'zeros' are not yet supported")

        return F.conv1d(input, weight, bias, self.stride, self.padding, self.dilation, self.groups)

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)

        return self._conv_forward(input, self.weight, self.bias)


class Conv2d(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, padding_mode='zeros', bias=True):
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = padding if isinstance(padding, str) else padding
        dilation = _pair(dilation)
        super().__init__(in_channels, out_channels, kernel_size, stride, padding,
                         dilation, False, _pair(0), groups, padding_mode, bias)

    def _conv_forward(self, input, weight, bias):
        if self.padding_mode != 'zeros':
            raise NotImplementedError("Padding modes other than 'zeros' are not yet supported")

        return F.conv2d(input, weight, bias, self.stride, self.padding, self.dilation, self.groups)

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)

        return self._conv_forward(input, self.weight, self.bias)


class Conv3d(_ConvNd):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1, padding_mode='zeros', bias=True):
        kernel_size = _triple(kernel_size)
        stride = _triple(stride)
        padding = padding if isinstance(padding, str) else padding
        dilation = _triple(dilation)
        super().__init__(in_channels, out_channels, kernel_size, stride, padding,
                         dilation, False, _triple(0), groups, padding_mode, bias)

    def _conv_forward(self, input, weight, bias):
        if self.padding_mode != 'zeros':
            raise NotImplementedError("Padding modes other than 'zeros' are not yet supported")

        return F.conv3d(input, weight, bias, self.stride, self.padding, self.dilation, self.groups)

    def forward(self, input):
        if not isinstance(input, Tensor):
            input = Tensor.create(input, requires_grad=self._training, device=self._device)

        return self._conv_forward(input, self.weight, self.bias)



import os
import pickle
import tarfile
from collections import OrderedDict

import numpy as np

import neuroscribe as ns
from neuroscribe.core._utils._data import (_decompress_file, _request_file,
                                           _save_file)
from neuroscribe.utils.data.transforms import Compose
from neuroscribe.utils.data.utils import read_data

__all__ = ['Dataset', 'MNIST', 'FashionMNIST']


class Dataset:
    def __init__(self):
        self._device = 'cpu'
        super().__setattr__('_parameters', OrderedDict())

    def __setattr__(self, name, value):
        if isinstance(value, list):
            for val in value:
                if name in self._parameters:
                    if isinstance(self._parameters[name], list):
                        self._parameters[name].append(ns.tensor(val))
                    else:
                        self._parameters[name] = [
                            self._parameters[name], ns.tensor(val)]
                else:
                    self._parameters[name] = ns.tensor(val)

        elif isinstance(value, np.ndarray):
            if '_parameters' not in self.__dict__:
                raise AttributeError(
                    "cannot assign parameters before Module.__init__() call")
            if name in self._parameters:
                if isinstance(self._parameters[name], list):
                    self._parameters[name].append(ns.tensor(value))
                else:
                    self._parameters[name] = [
                        self._parameters[name], ns.tensor(value)]
            else:
                self._parameters[name] = ns.tensor(value)
        else:
            super().__setattr__(name, value)

    def __getattr__(self, name):
        if '_parameters' in self.__dict__:
            _parameters = self.__dict__['_parameters']
            if name in _parameters:
                return _parameters[name]

        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'")

    def to(self, device):
        self._device = device
        for name, param in self._parameters.items():
            if isinstance(param, list):
                self._parameters[name] = [item.to(device) for item in param]
            else:
                self._parameters[name] = param.to(device)


class _MNISTDataset(Dataset):
    def __init__(self, images_file_path, labels_file_path, transform=None):
        super().__init__()
        self.images = read_data(images_file_path)
        self.labels = read_data(labels_file_path)
        self.transform = Compose(transform) if isinstance(
            transform, list) else transform

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        x = self.images[idx]
        y = self.labels[idx]

        if self.transform:
            x = ns.tensor(
                self.transform(x.asnumpy()),
                requires_grad=False,
                device=self._device
            )

        return x, y

    @staticmethod
    def _download_and_cache_mnist(root, name, base_url, urls):
        mnist_dir = os.path.join(root, name)
        os.makedirs(mnist_dir, exist_ok=True)
        for key, relative_path in urls.items():
            url = os.path.join(base_url, relative_path)
            file_name = relative_path.split('/')[-1]
            file_path = os.path.join(mnist_dir, file_name)
            decompressed_file_path = file_path[:-
                                               3] if file_path.endswith('.gz') else file_path

            if not os.path.exists(decompressed_file_path):
                if not os.path.exists(file_path):
                    response = _request_file(url)
                    _save_file(response, file_path)
                _decompress_file(file_path)

    @staticmethod
    def _get_dataset(root, dataset_name, base_url, urls, train, download, transform):
        if download:
            _MNISTDataset._download_and_cache_mnist(
                root, dataset_name, base_url, urls)

        subdir = 'MNIST' if dataset_name == 'mnist' else 'fashion_mnist'
        if train:
            images_file_path = os.path.join(
                root, subdir, "train-images-idx3-ubyte")
            labels_file_path = os.path.join(
                root, subdir, "train-labels-idx1-ubyte")
        else:
            images_file_path = os.path.join(
                root, subdir, "t10k-images-idx3-ubyte")
            labels_file_path = os.path.join(
                root, subdir, "t10k-labels-idx1-ubyte")

        return _MNISTDataset(images_file_path, labels_file_path, transform=transform)


class _CIFAR10Dataset(Dataset):
    def __init__(self, data, labels, transform=None):
        super().__init__()
        self.data = data
        self.labels = labels
        self.transform = Compose(transform) if isinstance(
            transform, list) else transform

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        x = self.data[idx.item()]
        y = self.labels[idx.item()]

        if self.transform:
            x = ns.tensor(
                self.transform(x.asnumpy()),
                requires_grad=False,
                device=self._device
            )

        return x, y

    @staticmethod
    def _download_and_cache_cifar10(root, base_url, file_name):
        cifar10_dir = os.path.join(root, 'cifar10')
        os.makedirs(cifar10_dir, exist_ok=True)
        file_path = os.path.join(cifar10_dir, file_name)

        if not os.path.exists(file_path):
            response = _request_file(base_url + file_name)
            _save_file(response, file_path)

        with tarfile.open(file_path, 'r:gz') as tar:
            tar.extractall(path=cifar10_dir)

    @staticmethod
    def _load_data_from_batch(file):
        with open(file, 'rb') as fo:
            dict = pickle.load(fo, encoding='bytes')
        return dict[b'data'], dict[b'labels']

    @staticmethod
    def _load_data(images_file_path, labels_file_path):
        with open(images_file_path, 'rb') as fo:
            data = np.load(fo, allow_pickle=True)
        with open(labels_file_path, 'rb') as fo:
            labels = np.load(fo, allow_pickle=True)
        data = data.reshape(-1, 3, 32, 32).astype(np.float32)
        return data, labels

    @staticmethod
    def _save_data_as_npy(images_file_path, labels_file_path, data, labels):
        np.save(images_file_path, data)
        np.save(labels_file_path, labels)

    @staticmethod
    def _get_dataset(root, dataset_name, base_url, file_name, train, download, transform):
        if download:
            _CIFAR10Dataset._download_and_cache_cifar10(
                root, base_url, file_name)

        cifar10_dir = os.path.join(root, 'cifar10')
        if train:
            data, labels = [], []
            for batch_num in range(1, 6):
                batch_file = os.path.join(
                    cifar10_dir, f'cifar-10-batches-py/data_batch_{batch_num}')
                batch_data, batch_labels = _CIFAR10Dataset._load_data_from_batch(
                    batch_file)
                data.append(batch_data)
                labels.append(batch_labels)
            data = np.vstack(data).reshape(-1, 3, 32, 32).astype(np.float32)
            labels = np.hstack(labels)
        else:
            test_batch_file = os.path.join(
                cifar10_dir, 'cifar-10-batches-py/test_batch')
            data, labels = _CIFAR10Dataset._load_data_from_batch(
                test_batch_file)
            data = data.reshape(-1, 3, 32, 32).astype(np.float32)

        images_file_path = os.path.join(cifar10_dir, 'images.npy')
        labels_file_path = os.path.join(cifar10_dir, 'labels.npy')
        _CIFAR10Dataset._save_data_as_npy(
            images_file_path, labels_file_path, data, labels)

        return _CIFAR10Dataset(data, labels, transform=transform)


def CIFAR10(root, train=True, download=True, transform=None):
    _BASE_URL = "https://www.cs.toronto.edu/~kriz/"
    _FILE_NAME = "cifar-10-python.tar.gz"
    return _CIFAR10Dataset._get_dataset(root, 'cifar10', _BASE_URL, _FILE_NAME, train, download, transform)


def MNIST(root, train=True, download=True, transform=None):
    _BASE_URL = "https://storage.googleapis.com/cvdf-datasets/mnist/"
    _URLS = {
        "train_images": "train-images-idx3-ubyte.gz",
        "train_labels": "train-labels-idx1-ubyte.gz",
        "test_images": "t10k-images-idx3-ubyte.gz",
        "test_labels": "t10k-labels-idx1-ubyte.gz",
    }
    return _MNISTDataset._get_dataset(root, 'mnist', _BASE_URL, _URLS, train, download, transform)


def FashionMNIST(root, train=True, download=True, transform=None):
    _BASE_URL = "https://github.com/zalandoresearch/fashion-mnist/raw/master/data/fashion/"
    _URLS = {
        "train_images": "train-images-idx3-ubyte.gz",
        "train_labels": "train-labels-idx1-ubyte.gz",
        "test_images": "t10k-images-idx3-ubyte.gz",
        "test_labels": "t10k-labels-idx1-ubyte.gz",
    }
    return _MNISTDataset._get_dataset(root, 'fashion_mnist', _BASE_URL, _URLS, train, download, transform)



import pickle
from collections import OrderedDict

from neuroscribe.core._tensor_lib._tensor import Tensor


def save(model, filename):
    if not isinstance(model, Module):
        raise TypeError(f"Cannot save object of type '{type(model).__name__}', expected 'Module'")
    filename = filename + '.nsm' if not filename.endswith('.nsm') else filename
    with open(filename, "wb") as f:
        pickle.dump(model, f)
    print(f"Module saved to '{filename}'")


def load(filename):
    filename = filename + '.nsm' if not filename.endswith('.nsm') else filename
    with open(filename, "rb") as f:
        module = pickle.load(f)
    return module


class Module:

    def __init__(self):

        super().__setattr__('_training', False)
        super().__setattr__('_device', 'cpu')
        super().__setattr__('_parameters', OrderedDict())
        super().__setattr__('_modules', OrderedDict())

    def __setattr__(self, name, value):
        if isinstance(value, Tensor):
            if '_parameters' not in self.__dict__:
                raise AttributeError("cannot assign parameters before Module.__init__() call")
            self._parameters[name] = value
        elif isinstance(value, Module):
            if '_modules' not in self.__dict__:
                raise AttributeError("cannot assign module before Module.__init__() call")
            self._modules[name] = value
        else:
            super().__setattr__(name, value)

    def __getattr__(self, name):
        if '_parameters' in self.__dict__:
            _parameters = self.__dict__['_parameters']
            if name in _parameters:
                return _parameters[name]
        if '_modules' in self.__dict__:
            _modules = self.__dict__['_modules']
            if name in _modules:
                return _modules[name]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def __call__(self, *input):
        return self.forward(*input)

    def forward(self, *input): raise NotImplementedError

    def parameters(self):
        for param in self._parameters.values():
            yield param
        for module in self._modules.values():
            yield from module.parameters()

    def add_parameter(self, name, parameter):

        if '_parameters' not in self.__dict__:
            raise AttributeError("cannot assign parameters before Module.__init__() call")

        if not isinstance(parameter, Tensor) and parameter is not None:
            raise TypeError(f"parameter must be of type 'Tensor'. Got {type(parameter)}")
        elif parameter is None:
            raise ValueError("parameter cannot be None")
        elif not isinstance(name, str):
            raise TypeError(f"parameter name should be a string. Got {type(name)} instead")
        elif hasattr(self, name) and name not in self._parameters:
            raise KeyError(f"attribute '{name}' already exists")
        elif '.' in name:
            raise KeyError("parameter name cannot contain \".\"")
        elif name == '':
            raise KeyError("parameter name cannot be an empty string \"\"")

        self._parameters[name] = parameter

    def add_module(self, name, module):

        if '_modules' not in self.__dict__:
            raise AttributeError("cannot assign module before Module.__init__() call")

        if not isinstance(module, Module) and module is not None:
            raise TypeError(f"{type(module)} is not a Module subclass. Got {type(module)}")
        elif module is None:
            raise ValueError("module cannot be None")
        elif not isinstance(name, str):
            raise TypeError(f"module name should be a string. Got {type(name)} instead")
        elif hasattr(self, name) and name not in self._modules:
            raise KeyError(f"attribute '{name}' already exists")
        elif '.' in name:
            raise KeyError("module name cannot contain \".\"")
        elif name == '':
            raise KeyError("module name cannot be an empty string \"\"")

        self._modules[name] = module

    def to(self, device):
        self._device = device
        for name, param in self._parameters.items():
            self._parameters[name] = param.to(device)
        for module in self._modules.values():
            module.to(device)

    def _mode(self, *, training):
        self._training = training
        for param in self._parameters.values():
            param.requires_grad = training
        for module in self._modules.values():
            module._mode(training=training)

    def train(self):
        self._mode(training=True)

    def eval(self):
        self._mode(training=False)
